package Parsing.LineParsers;

import Parsing.DBPrimitives.FixedStrTokens.AbsTaBle;
import PrologDB.Column;
import PrologDB.DB;
import PrologDB.DBSchema;
import PrologDB.SubTableSchema;
import PrologDB.Table;
import PrologDB.TableSchema;
import PrologDB.Tuple;
import PrologDB.Schema;
import Parsing.DBPrimitives.Patterns.DBaseStmt;
import Parsing.DBPrimitives.Patterns.OODBStmt;
import Parsing.DBPrimitives.Patterns.SubTableStmt;
import Parsing.DBPrimitives.Patterns.TupleStmt;
import Parsing.DBPrimitives.Patterns.XTableStmt;
import Parsing.TBPrims.SaveToken.Option;
import Parsing.TBPrims.SaveToken.TaBleModifier;
import PrologDB.ErrorReport;
import PrologDB.OOSchema;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

/**
 * this is a simple line-parser for database and schema specifications Generally
 * this class is internal to PrologDB package and should not be called
 * externally
 */
public class DBLineParsers {

    /**
     * turn Debug on if you want to see stack trace on Errors
     */
    static boolean Debug = false;

    /**
     * a dbase declaration is on a single line. this method parses the given
     * line @ lineno returning the DBSchema given declared SubTableSchema, also
     * returns a list of tableNames (not their schemas). ParseExceptions thrown
     * if unable to parse
     *
     * @param line -- a string
     * @param lineno -- the line number associated with this line
     * @param tableNames -- an empty list which is to be filled with table names
     * and returned
     * @return DBSchema object with SubTableSchema of schema initialized; also
     * returns list of tableNames -- to be used later in parsing
     */
    public static DBSchema parseDBaseLine(String line, int lineno, List<String> tableNames) {
        // Step 1: parse the line
        LineToParse l = new LineToParse(line, lineno);
        DBaseStmt st = new DBaseStmt(l);
        st.parse();

        // Step 2: unpack parsing (schemaName, list-of-tables)
        LinkedList<String> list = l.parseList;
        String schemaName = list.removeFirst();
        DBSchema dbs = new DBSchema(schemaName);
        tableNames.addAll(list);
        return dbs;
        // note: no subtable declarations are parsed.  So the DBSchema is unfinished
    }

    /**
     * a oodbase declaration is on a single line. this method parses the given
     * line @ lineno returning the DBSchema given declared SubTableSchema, also
     * returns a list of tableNames (not their schemas). ParseExceptions thrown
     * if unable to parse
     *
     * @param line -- a string
     * @param lineno -- the line number associated with this line
     * @param tableNames -- an empty list which is to be filled with table names
     * and returned
     * @return DBSchema object with SubTableSchema of schema initialized; also
     * returns list of tableNames -- to be used later in parsing
     */
    public static OOSchema parseOODBLine(String line, int lineno, List<String> tableNames) {
        // Step 1: parse the line
        LineToParse l = new LineToParse(line, lineno);
        OODBStmt st = new OODBStmt(l);
        st.parse();

        // Step 2: unpack parsing (schemaName, list-of-tables)
        LinkedList<String> list = l.parseList;
        String schemaName = list.removeFirst();
        OOSchema dbs = new OOSchema(schemaName);
        tableNames.addAll(list);
        return dbs;
        // note: no subtable declarations are parsed.  So the DBSchema is unfinished
    }

    /**
     * a subtable declaration is on a single line. this method parses the given
     * line at lineno and adds the subtable declaration to DBSchema dbs. A
     * ParseException is thrown if something wrong (parse or semantic error) is
     * discovered.
     *
     * @param line to parse
     * @param lineno of line
     * @param dbs to add subtable definition
     * @param filename name of file from which this line was taken
     * @param er error reporter
     */
    public static void parseSubTableLine(String line, int lineno, Schema dbs, String filename, ErrorReport er) {
        if (dbs == null) {
            er.add("line %d: subtable declaration before db declaration", lineno);
            return;
        }

        // Step 1: parse the line
        LineToParse l = new LineToParse(line, lineno);
        SubTableStmt st = new SubTableStmt(l);
        try {
            st.parse();
        } catch (RuntimeException e) {
            er.add("line %d: error in parsing subtable decl: %s", lineno, e.getMessage());
            return;
        }

        // Step 2: unpack parsing to get table SubTableSchema, and list of values, and create an empty tuple
        LinkedList<String> list = l.parseList;
        String tableName = list.removeFirst();  // this is the super table SubTableSchema
        TableSchema suptab;
        suptab = dbs.getTableSchema(tableName);
        if (suptab == null) {
            er.add("line %d: in subtable decl: table %s was not yet declared in db schema", lineno, tableName);
            return;
        }
        ArrayList<TableSchema> subTableSchemas = new ArrayList<>();
        for (String name : list) {
            TableSchema ts = dbs.getTableSchema(name);
            if (ts == null) {
                er.add("line %d: subtable %s of supertable %s does not exist", lineno, name, tableName);
                continue;
            }
            subTableSchemas.add(ts);
        }

        // Step 3: assemble the subtableschema declaaration
        SubTableSchema sts = new SubTableSchema(suptab, subTableSchemas);
        dbs.add(sts);
    }


    /**
     * parses a line at lineno, expecting to parse a legal prolog tuple
     * declaration. The extracted tuple is inserted into database db; the
     * possible table declarations are defined in db.getTableSchema() of the
     * database
     *
     * @param line to parse
     * @param lineno line number of the line
     * @param db database in which to insert
     * @param fileName name of file from which this line was read
     * @param er error reporter
     *
     */
    public static void parseTupleLine(String line, int lineno, DB db, String fileName, ErrorReport er) {
        // Step 1: parse tuple
        LineToParse l = new LineToParse(line, lineno);
        TupleStmt st = new TupleStmt(l);
        try {
            st.parse();
        } catch (RuntimeException e) {
            er.add(e.getMessage());
            return;
        }

        // Step 2: unpack parsing to get table and list of values, and create an empty tuple
        LinkedList<String> list = l.parseList;
        String tableName = list.removeFirst(); // this is the table SubTableSchema
        Table tbl = db.getTable(tableName);
        if (tbl == null) {
            er.add("line %d: table %s undefined", lineno, tableName);
            return;
        }
        TableSchema tableDef = db.getTableSchema(tableName);
        Tuple newTuple = new Tuple(tableDef);
        if (list.size() != tableDef.size()) {
            er.add("line %d: tuple for table %s has incorrect number of values", lineno, tableDef.getName());
            return;
        }
        // Step 3: assemble the tuple
        for (int i = 0; i < tableDef.size(); i++) {
            String value = list.removeFirst(); // this is a column value
            if (value.startsWith("'") || value.startsWith("\"")) {
                value = value.substring(1, value.length() - 1);
            }
            Column target = tableDef.getColumn(i);
            String columnName = target.getName().replace("\"", "");
            newTuple.set(columnName, value);
        }
        // Step 4: add the tuple to the table
        tbl.add(newTuple);
    }

    /**
     * ******************************************************************************************
     */
    // TB is the next generation in MDELite schema specficiations.
    // This is the new, next-generation ParseTableLine method
    /**
     * a table declaration is on a single line. this method parses the given
     * line @ lineno and adds the table definition to schema dbs
     *
     * @param line to parse
     * @param lineno of line
     * @param dbs to add table declaration
     * @param filename name of file from which this line was read
     * @param tableNames --list of legal table names
     * @param er -- error reporter
     */
    public static void parseTableLine(String line, int lineno, Schema dbs, String filename, List<String> tableNames, ErrorReport er) {
        // Step 0: simple error checking;
        if (dbs == null) {
            er.add("line %d: table declared before db schema", lineno);
            return;
        }
        LineToParse l = new LineToParse(line, lineno);
        TableSchema ts = DBLineParsers.parseTableLine(l, er);
        if (ts == null) {
            return;
        }
        if (tableNames!=null && !tableNames.contains(ts.getName())) {
            er.add("line %d: table %s undeclared in db schema declaration", lineno, ts.getName());
            return;
        }
        if (dbs.getTableSchema(ts.getName()) != null) {
            er.add("line %d: table %s declared multiple times", lineno, ts.getName());
            return;
        }
        dbs.add(ts);
    }

    /**
     * a table declaration is on a single line. this method parses the given
     * line @ lineno; returns tableschema
     *
     * @param l -- LineToParse
     * @param er error reporter
     * @return TableSchema of line input
     */
    public static TableSchema parseTableLine(LineToParse l, ErrorReport er) {
        // Step 1: parse line
        XTableStmt st = new XTableStmt(l);
        try {
            st.parse();
        } catch (RuntimeException e) {
            er.add(e.getMessage());
            return null;
        }
        // Step 2: unpack parsing to get table name + (name, option/table, type)+
        LinkedList<String> list = l.parseList;
        String WhatAmI = list.removeFirst(); // table or abs_table
        String tableName = list.removeFirst(); // this is the table name
        TableSchema tableDef = new TableSchema(tableName);
        if (WhatAmI.equals(AbsTaBle.tok)) {
            tableDef.makeAbstract();
        }
        while (list.isEmpty() == false) {
            String name = list.removeFirst(); // column name
            String opt = list.removeFirst(); // opt = (empty) | table | optional
            Column.opts o;
            switch (opt) {
                case TaBleModifier.tok:
                    o = Column.opts.table;
                    break;
                case Option.tok:
                    o = Column.opts.option;
                    break;
                default:
                    o = Column.opts.empty;
                    break;
            }
            String kind = list.removeFirst(); // kind = typename
            boolean isQuoted = name.startsWith("\"") || name.startsWith("'");
            String fleased = name.replaceAll("\"", "").replaceAll("'", "");
            Column newColumn = new Column(fleased, o, kind, isQuoted);
            tableDef.add(newColumn);
        }
        return tableDef;
    }
}
