package PrologDB;

import MDLUtilities.Error;
import Parsing.DBPrimitives.FixedStrTokens.TaBle;
import Parsing.TBPrims.SaveToken.Option;

/**
 * A column is a read-only, non-modifiable (String name, boolean isQuoted) pair
 */
public class Column {

    public enum opts {
        option(Option.tok + " "), table(TaBle.tok + " "), empty("");

        private final String msg;

        /**
         * required constructor
         *
         * @param msg -- string interpretation of error
         */
        opts(String msg) {
            this.msg = msg;
        }

        /**
         * @return message of error
         */
        public String getMsg() {
            return msg;
        }

        public static opts toEnum(String s) {
            for (opts o : opts.values()) {
                if (o.msg.startsWith(s)) {
                    return o;
                }
            }
            return null;
        }

        public static opts toEnumEH(String s) {
            opts result = toEnum(s);
            if (result == null) {
                throw new RuntimeException("Unknown opt enum " + s);
            }
            return result;
        }
    }

    /**
     * State of a column name -- name of column isQuoted -- does the column
     * contain single-quoted values? and kind
     */
    private final String name;
    private final boolean isQuoted;
    private final String kind;  // Table name whose ids populate this column
    private final opts opt;

    // CONSTRUCTOR
    /**
     * copy Column constructor
     *
     * @param c -- column to copy
     */
    public Column(Column c) {
        name = c.name;
        isQuoted = c.isQuoted;
        kind = c.kind;
        opt = opts.empty;
    }

    /**
     * rename constructor
     *
     * @param c -- column to copy
     * @param newName -- new column name
     */
    public Column(Column c, String newName) {
        name = newName;
        isQuoted = c.isQuoted;
        kind = c.kind;
        opt = opts.empty;
    }

    public static Column[] New(String... names) {
        Column[] array = new Column[names.length];
        int i = 0;
        for (String n : names) {
            array[i++] = new Column(n);
        }
        return array;
    }

    /**
     * if qname is a single-quoted string (or if you're lazy, if first char of
     * qname is '), then column is quoted
     *
     * @param qname - quoted or unquoted name
     */
    public Column(String qname) {
        String[] array = qname.trim().split(":");
        String cname = array[0];
        if (cname.equals("")) {
            throw Error.toss(Error.columnFormatBad, qname);
        }
        if (array.length > 2) {
            throw Error.toss(Error.columnFormatBad, qname);
        }

        if (cname.startsWith("'") | cname.startsWith("\"")) {
            this.name = cname.replace("'", "").replace("\"", "");
            this.isQuoted = true;
            kind = "qstr";
        } else {
            this.name = cname;
            this.isQuoted = false;
            if (array.length == 2) {
                kind = array[1];
            } else {
                kind = "str";
            }
        }
        opt = opts.empty;
    }

    public Column(String name, opts opt, String kind, boolean isQuoted) {
        this.name = name;
        this.opt = opt;
        this.kind = kind;
        this.isQuoted = isQuoted;
    }

// ACCESSOR AND INFO
    /**
     *
     * @return string name of column
     */
    public String getName() {
        return name;
    }

    /**
     *
     * @return true if column values are single quoted
     */
    public boolean isQuoted() {
        return isQuoted;
    }

    /**
     * @return opts enum
     */
    public opts getOpt() {
        return opt;
    }

    /**
     * return kind
     *
     * @return kind
     */
    public String getKind() {
        return kind;
    }
    
    /**
     * 
     * @return true if column has a primitive type
     */
    public boolean isPrimType() {
        for (PrimType p : PrimType.values()) {
            if (p.getName().equals(kind))
                return true;
        }
        return false;
    }

    // UTILITIES
    /**
     * returns true if 'this' column equals the given otherColumn. Column
     * equality means both columns have the same name, same isQuoted property,
     * same kind and modifier.
     *
     * @param otherColumn that should be equal to 'this' column
     * @return true if 'this' column equals the given otherColumn.
     */
    public boolean equals(Column otherColumn) {
        if (otherColumn == null) {
            return false;  // nothing to compare
        }
        String tempName = otherColumn.getName();
        if (!name.equals(tempName)) {
            return false;  // names do not match
        }
        if (isQuoted != otherColumn.isQuoted()) {
            return false; // quotation property does not match
        }
        String kname = otherColumn.kind;
        if (!kind.equals(kname)) {
            return false;
        }
        if (opt != otherColumn.opt) {
            return false;
        }
        return true;  // they match!
    }

    // PRINT
    /**
     * a String schema specification of this column is its name in/not-in double
     * quotes
     *
     * @return a String of a schema specification for 'this' column
     */
    @Override
    public String toString() {
        String q = isQuoted ? "\"" : "";
        boolean isString = kind.equals(PrimType.QStr.getName())
                || kind.equals(PrimType.Str.getName());
        if (isString && opt == opts.empty) {
            return q + name + q;
        }
        if (kind.equals("")) {
            System.out.println("shouldn't be here");
        }
        String result = String.format("%s%s%s:%s%s", q, name, q, opt.getMsg(), kind);
        return result;
    }
    
}
