package PrologDB;

import MDLUtilities.Error;
import MDLUtilities.ParsE;
import Parsing.DBPrimitives.FixedStrTokens.AbsTaBle;
import Parsing.DBPrimitives.FixedStrTokens.DBase;
import Parsing.DBPrimitives.FixedStrTokens.OODB;
import Parsing.DBPrimitives.FixedStrTokens.SubTaBle;
import Parsing.DBPrimitives.FixedStrTokens.TaBle;
import Parsing.LineParsers.DBLineParsers;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;

/**
 * A DBSchema object defines the schema of a database
 *
 */
public class DBSchema extends Schema {
    // CONSTRUCTORS

    /**
     * This constructor is used for the incremental construction of database
     * tableSchemas. First provide a SubTableSchema of the schema, then
     * incrementally add tableSchemas per table and then subtables. By default,
     * schema is not flattened.
     *
     * @param name -- SubTableSchema given to the DBSchema
     */
    public DBSchema(String name) {
        super(name);
    }

    /**
     * copyFrom a database schema. Only unless it is unfinished can the copyFrom
     * be unfinished.
     *
     * @return copied schema
     */
    public DBSchema copy() {
        DBSchema s = new DBSchema(name);
        s.copyFrom(this);
        s.finished = finished;
        return s;
    }

    /**
     * reads .ooschema and .schema files; throws Error when an error is
     * encountered. Returns DBSchema declaration in file; may flatten it if
     * .ooschema is read.
     *
     * @param schemafile -- SubTableSchema of schema file
     * @return DBSchema declaration of file
     */
    public static DBSchema read(String schemafile) {
        File in = new File(schemafile);
        return DBSchema.read(in);
    }

    /**
     * another version of readDBSchema
     *
     * @param schemafile -- File to read
     * @return DBSchema object
     */
    public static DBSchema read(File schemafile) {
        String filename = schemafile.getPath().replace("\\", "/");
        LineNumberReader br = readFileCheck(schemafile, filename);
        return read(filename, br);
    }

    /**
     * helper method -- does basic checks on File schemafile
     *
     * @param schemafile -- file that contains database schema
     */
    private static LineNumberReader readFileCheck(File schemafile, String filename) {
        LineNumberReader br;

        if (filename.endsWith(".ooschema.pl") || filename.endsWith(".pl")) {
            try {
                br = new LineNumberReader(new InputStreamReader(new FileInputStream(filename)));
                return br;
            } catch (FileNotFoundException e) {
                throw Error.toss(Error.schemaFileNoExist, filename);
            }
        }
        throw Error.toss(Error.wrongFileName, filename);
    }

    /**
     * used when it is easier to pass in a string declaration of a DBSchema
     * rather than building it programmatically
     *
     * @param nameOfSchema -- name of pseudo file
     * @param stringDefOfSchema -- string contents (typically multiline) of
     * pseudo file
     * @return DBSchema -- of pseudo file
     */
    public static DBSchema read(String nameOfSchema, String stringDefOfSchema) {
        InputStream is = new ByteArrayInputStream(stringDefOfSchema.getBytes());
        LineNumberReader br = new LineNumberReader(new InputStreamReader(is));
        return read(nameOfSchema, br);
    }

    /**
     * reads prolog schema from File schemafile; can read .ooschema and .schema
     * files
     *
     * @param filename the value of filename
     * @param br - line number reader of file to read
     * @return the PrologDB.DBSchema
     */
    private static DBSchema read(String filename, LineNumberReader br) {
        DBSchema dbs = null;
        List<String> tableNames = new ArrayList<>();
        ErrorReport er = new ErrorReport();
        try {
            String line;
            while ((line = br.readLine()) != null) {
                line = line.trim();
                if (line.equals("")
                        || line.startsWith("%")
                        || line.startsWith("//")) {
                    continue;
                }
                int lineno = br.getLineNumber();

                if (line.startsWith(DBase.tok)) {
                    if (dbs != null) {
                        er.add("line %d: more than one %s declaration", DBase.tok);
                        continue;
                    }
                    dbs = DBLineParsers.parseDBaseLine(line, lineno, tableNames);
                } else if (line.startsWith(TaBle.tok) || line.startsWith(AbsTaBle.tok)) {
                    DBLineParsers.parseTableLine(line, lineno, dbs, filename, tableNames, er);
                } else if (line.startsWith(SubTaBle.tok)) {
                    DBLineParsers.parseSubTableLine(line, lineno, dbs, filename, er);
                } else if (line.startsWith(OODB.tok)) {
                    throw ParsE.toss(ParsE.ooInappropriate, lineno);
                }
                // skip anything else -- could be data

            }
            br.close();
            // we have parsed a database schema file
            // Let's make sure that the tables declared match what is in the dbase declaration

            if (dbs != null) {
                dbs.finished = !filename.endsWith(".ooschema.pl");

                // now basic verification.  tableNames is a list of names
                // produced when parsing a dbase declaration.  make sure that
                // schemas for these tables (and no others) have been created
                Schema.basicVerification(dbs, tableNames, er);
            } else if (!er.hasErrors()) {
                throw Error.toss(Error.schemaNotValid, filename);
            }

        } catch (IOException e) {
            throw Error.toss(Error.ioerror, filename, e.getMessage());
        }
        er.printEH(System.out);
        return dbs;
    }

    /**
     * full (file) SubTableSchema of a schema is (SubTableSchema).schema.pl
     *
     * @return manufactured file SubTableSchema of schema
     */
    public String getFullName() {
        return name + ".schema.pl";
    }

    // PRINT UTILITIES
    String printDBaseDecl() {
        return SchemaDeclString(DBase.tok);
    }

    /**
     * converts DBSchema to a string. Works for ooschema and schema files
     *
     * @return print string of DBSchema
     */
    public String toString() {
        String result = String.format("%s\n", printDBaseDecl());
        for (TableSchema ts : tableSchemas) {
            result = result + ts.toString();
        }
        if (!subtables.isEmpty()) {
            result = result + "\n";
            for (SubTableSchema st : subtables) {
                result = result + st.toString();
            }
        }
        return result;
    }

    /**
     * print the database schema definition to PrintStream out; typically used
     * for demonstrations as it doesn't check the format of the filename; Use
     * print() or print(String).
     *
     * @param out PrintStream destination
     */
    @Override
    public void print(PrintStream out) {
        out.print(this.toString());
    }

    /**
     * prints to System.out
     */
    @Override
    public void print() {
        print(System.out);
    }

    /**
     * print DBSchema in file with filename
     *
     * @param filename -- name of file in which to print DBSchema
     */
    @Override
    public void print(String filename) {
        print(filename, ".schema.pl");
    }
}
