package PrologDB;

import Parsing.LineParsers.*;
import MDLUtilities.Error;
import MDLUtilities.ParsE;
import Parsing.DBPrimitives.FixedStrTokens.AbsTaBle;
import Parsing.DBPrimitives.FixedStrTokens.DBase;
import Parsing.DBPrimitives.FixedStrTokens.OODB;
import Parsing.DBPrimitives.FixedStrTokens.TaBle;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

/**
 * A DBSchema object defines the schema of a database
 *
 */
public class OOSchema extends Schema {

    // CONSTRUCTORS
    /**
     * This constructor is used for the incremental construction of dbase or
     * oodb Schemas. Can incrementally add tables and then add subtable schemas.
     *
     *
     * @param name -- name given to schema
     */
    public OOSchema(String name) {
        super(name);
        this.name = name;
    }

    /**
     * copyFrom an oodb schema.
     *
     * @return copied schema
     */
    public OOSchema copy() {
        OOSchema s = new OOSchema(name);
        s.copyFrom(this);
        return s;
    }
    
    // ACCESSORS, SETTERS, and INFORMATION
    /**
     * full (file) SubTableSchema of a schema is (SubTableSchema).schema.pl
     *
     * @return manufactured file SubTableSchema of schema
     */
    @Override
    public String getFullName() {
        return name + ".ooschema.pl";
    }
    
    /** 
     * declare OOschema finished -- can no longer be edited.
     */
    public void setFinished() {
        finished = true;
    }

    // READ FILE UTILITIES
    /**
     * reads .ooschema; throws Error when an error is encountered. Returns
     * DBSchema declaration in file; may flatten it if .ooschema is read.
     *
     * @param schemafile -- SubTableSchema of schema file
     * @return DBSchema declaration of file
     */
    public static OOSchema read(String schemafile) {
        File in = new File(schemafile);
        return OOSchema.read(in);
    }

    /**
     * another version of readSchema
     *
     * @param schemafile -- File to read
     * @return DBSchema object
     */
    public static OOSchema read(File schemafile) {
        String filename = schemafile.getPath().replace("\\", "/");
        LineNumberReader br = readFileCheck(schemafile, filename);
        return read(filename, br);
    }

    /**
     * helper method -- does basic checks on File schemafile
     *
     * @param schemafile -- file that contains database schema
     */
    private static LineNumberReader readFileCheck(File schemafile, String filename) {
        LineNumberReader br;

        if (filename.endsWith(".ooschema.pl")) {
            try {
                br = new LineNumberReader(new InputStreamReader(new FileInputStream(filename)));
                return br;
            } catch (FileNotFoundException e) {
                throw Error.toss(Error.schemaFileNoExist, filename);
            }
        }
        throw Error.toss(Error.wrongFileName, filename);
    }

    /**
     * used when it is easier to pass in a string declaration of a DBSchema
     * rather than building it programmatically
     *
     * @param nameOfSchema -- name of pseudo file
     * @param stringDefOfSchema -- string contents (typically multiline) of
     * pseudo file
     * @return DBSchema -- of pseudo file
     */
    public static OOSchema read(String nameOfSchema, String stringDefOfSchema) {
        InputStream is = new ByteArrayInputStream(stringDefOfSchema.getBytes());
        LineNumberReader br = new LineNumberReader(new InputStreamReader(is));
        return read(nameOfSchema, br);
    }

    /**
     * reads prolog schema from File schemafile; can read .ooschema and .schema
     * files
     *
     * @param filename the value of filename
     * @param br - line number reader of file to read
     * @return the PrologDB.DBSchema
     */
    private static OOSchema read(String filename, LineNumberReader br) {
        OOSchema dbs = null;
        List<String> tableNames = new ArrayList<>();
        ErrorReport er = new ErrorReport();
        try {
            String line;
            while ((line = br.readLine()) != null) {
                line = line.trim();
                if (line.equals("")
                        || line.startsWith("%")
                        || line.startsWith("//")) {
                    continue;
                }
                int lineno = br.getLineNumber();

                try {
                    if (line.startsWith(OODB.tok)) {
                        if (dbs != null) {
                            throw ParsE.toss(ParsE.dbDeclsTooMany, lineno, filename);
                        }
                        dbs = DBLineParsers.parseOODBLine(line, lineno, tableNames);
                    } else if (line.startsWith(TaBle.tok) || line.startsWith(AbsTaBle.tok)) {
                        DBLineParsers.parseTableLine(line, lineno, dbs, filename,tableNames,er);
                    } else if (line.startsWith("subtable")) {
                        DBLineParsers.parseSubTableLine(line, lineno, dbs, filename, er);
                    } else if (line.startsWith(DBase.tok)) {
                        er.add("line %d: inappropriate to have a %s statement in an %s file", lineno, TaBle.tok, OODB.tok);
                    }
                    // ignore everything else
                } catch (RuntimeException e) {
                    throw ParsE.toss(ParsE.tableSchemaParseError, br.getLineNumber(), filename, e.getMessage());
                }
            }
            br.close();
            // we have parsed a database schema file
            // Let's make sure that the tables declared match what is in the dbase declaration

//            if (dbs == null) {
//                throw Error.toss(Error.schemaNotValid, filename);
//            }

            // now basic verification.  tableNames is a list of names
            // produced when parsing a dbase declaration.  make sure that
            // schemas for these tables (and no others) have been created
            if (dbs != null)
                basicVerification(dbs, tableNames,er);
            else
                er.add("no %s statement in this ooschema",OODB.tok );
        } catch (IOException e) {
            throw Error.toss(Error.ioerror, filename, e.getMessage());
        }
        er.printEH(System.out);
        return dbs;

        // finally, depending on who is calling this, the flatten attribute is to be set.
    }

    @Override
    public void print(String filename) {
        print(filename, ".ooschema.pl");
    }
    
    /**
     * @return the OOSchema declaration as a string
     */
    @Override
    public String toString() {
        return toString(OODB.tok);
    }
    
    /**
     * sets DB Schema ready for instantiation; flattens the schema if not
     * already flattened.
     * @return converted DBSchema object
     */
    public DBSchema finishAndPropagateAttributes() {
        if (!subtables.isEmpty()) {
            flatten();
        }
        return finishedAndIPropagatedAttributesMyself();
    }


    // UTILITIES
    /**
     *
     * @param list
     * @return
     */
    private SubTableSchema dominant(LinkedList<SubTableSchema> list) {
        boolean foundit = true;
        SubTableSchema last = null;

        for (SubTableSchema outer : list) {
            last = outer;
            for (SubTableSchema inner : list) {
                if (inner.contains(outer.getSuper())) {
                    foundit = false;
                    break;
                }
            }
            if (foundit) {
                return outer;
            }
            foundit = true;
        }
        throw Error.toss(Error.SchemaInheritanceCycle, last.getName());
    }
    
    /**
     * propagate attributes of superTables to subTables (operation is 
     * performed on schema defs)
     */
    private void flatten() {
        // Step 1: must order subtable list by dominance 
        //         (dbSchema dominates a child)
        LinkedList<SubTableSchema> list = subtables;
        subtables = new LinkedList<>();
        while (!list.isEmpty()) {
            SubTableSchema next = dominant(list);
            list.remove(next);
            subtables.add(next);
        }

        // Step 2: propagate dbSchema attributes to subattributes
        for (SubTableSchema st : subtables) {
            st.flatten();
        }
    }
}
