package PrologDB;

import MDLUtilities.Error;
import java.io.IOException;
import java.io.PrintStream;
import java.util.LinkedList;
import java.util.List;

public abstract class Schema {

    protected String name;
    protected final LinkedList<TableSchema> tableSchemas;
    protected LinkedList<SubTableSchema> subtables;

    protected boolean finished; // flattened sets finished in DBSchemas,
    // setFinished in OOSchemas.

    /**
     * This constructor is used for the incremental construction of dbase or
     * oodb Schemas. Can incrementally add tables and then add subtable schemas.
     *
     *
     * @param name -- name given to schema
     */
    protected Schema(String name) {
        this.name = name;
        this.tableSchemas = new LinkedList<>();
        subtables = new LinkedList<>();
        finished = false;
    }

    // SETTERS,GETTERS, INFO methods
    /**
     * full (file) SubTableSchema of a schema is (SubTableSchema).schema.pl
     *
     * @return manufactured file SubTableSchema of schema
     */
    public abstract String getFullName();

    /**
     * return number of TableSchemas (equivalently the number of Tables in a
     * schema instance) in this dbschema
     *
     * @return number of TableSchemas in this db schema
     */
    public int size() {
        return tableSchemas.size();
    }

    /**
     * copyFrom core elements of a database or oodb schema
     *
     * @param s -- copy TableSchemas and SubTables from s into 'this'
     */
    public void copyFrom(Schema s) {
        tableSchemas.addAll(s.tableSchemas);
        subtables.addAll(s.subtables);
//        associations.addAll(s.associations);
        finished = s.finished;
    }

    /**
     * @return the name of the schema
     */
    public String getName() {
        return name;
    }

    /**
     * rename Schema
     *
     * @param name new Schema Name
     */
    public void rename(String name) {
        this.name = name;
    }

    /**
     * @return true if the schema is no longer changeable and for dbschemas that
     * its supertable attributes have been propagated to its subtables.
     */
    public boolean isFinished() {
        return finished;
    }

    // UTILITIES
    public static void basicVerification(Schema dbs, List<String> tableNames, ErrorReport er) throws RuntimeException {
        // now basic verification.  tableNames is a list of names
        // produced when parsing a db declaration.  make sure that
        // schemas for these tables (and no others) have been created
        for (String s : tableNames) {
            if (dbs.getTableSchema(s) == null) {
                er.add("table schema %s undefined", s);
            } // throw error if not found
        }
        for (TableSchema ts : dbs.getTableSchemas()) {
            if (!tableNames.contains(ts.getName())) {
                er.add("table schema %s undefined", ts.getName());
            }
        }
    }

    /**
     * two Schemas are equal iff (a) they have the same set of table schemas,
     * and (b) they have the same set of subtableschemas; if so true is
     * returned.
     *
     * @param schema -- the other schema
     * @return true if both DBSchemas are equal.
     */
    public boolean equals(Schema schema) {
        if (!schema.getName().equals(getName())) {
            return false; // schema names are not the same
        }
        if (schema.size() != size()) {
            return false; // # of table definitions are not the same
        }
        for (TableSchema s : schema.getTableSchemas()) {
            TableSchema r = getTableSchema(s.getName());
            if (r == null || !r.equals(s)) {
                return false; // tables don't match
            }
        }
        if (subtables.size() != schema.subtables.size()) {
            return false;
        }
        for (SubTableSchema sts : schema.subtables) {
            SubTableSchema s2 = getSubTableSchema(sts.getName());
            if (!sts.equals(s2)) {
                return false; // have different subtable entries
            }
        }
        return true;
    }

    /**
     * two DBSchemas are equal iff (a) they have the same SubTableSchema, (b)
     * they have the same set of table schemas, and (c) they have the same set
     * of subtable schemas; if so this method returns. Otherwise a Error is
     * thrown
     *
     * @param schema.
     */
    public void equalsEH(Schema schema) {
        if (!schema.getName().equals(getName())) {
            throw Error.toss(Error.tableSchemaNamesDifferent, schema.getName(), getName());
        }
        if (schema.size() != size()) {
            throw Error.toss(Error.numberTablesDiffer, schema.getName(), getName());
        }
        for (TableSchema s : schema.getTableSchemas()) {
            TableSchema r = getTableSchema(s.getName());
            if (r == null) {
                throw Error.toss(Error.tableNoExist, s.getName(), schema.getName());
            }
            r.equalsEH(s);
        }
        if (subtables.size() != schema.subtables.size()) {
            throw Error.toss(Error.numberSubTablesDiffer, schema.getName(), getName());
        }
        for (SubTableSchema sts : schema.subtables) {
            SubTableSchema s2 = getSubTableSchema(sts.getName());
            sts.equalsEH(s2);
        }
    }

    // TABLE SCHEMA METHODS
    /**
     * add table schema tableschema to 'this' database schema
     *
     * @param tableschema to add to 'this' database schema
     */
    public void add(TableSchema tableschema) {
        String tableSchemaName = tableschema.getName();
        if (finished) {
            throw Error.toss(Error.SchemaNoMoreTables, tableSchemaName, name);
        }
        if (getTableSchema(tableSchemaName) != null) {
            throw Error.toss(Error.SchemaMultipleTableDecls, tableSchemaName);
        }
        tableSchemas.add(tableschema);
        tableschema.belongsToSchema = this;
    }

    /**
     * for adding schemas in bulk
     *
     * @param schemas -- list of table schemas
     */
    public void add(TableSchema... schemas) {
        for (TableSchema s : schemas) {
            Schema.this.add(s);
        }
    }

    /**
     * add subtable schema sts to 'this' database schema
     *
     * @param sts subtableschema to add to 'this' database
     */
    public void add(SubTableSchema sts) {
        String tableName = sts.getName();
        if (finished) {
            throw Error.toss(Error.SchemaNoMoreSubTables, name);
        }
        if (getSubTableSchema(tableName) != null) {
            throw Error.toss(Error.SchemaMultipleSubTableDecls, tableName);
        }
        subtables.add(sts);
        sts.belongstoSchema = this;
    }

    /**
     * add subtableschema in bulk
     *
     * @param sts -- an array of subtable schemas
     */
    public void add(SubTableSchema... sts) {
        for (SubTableSchema s : sts) {
            Schema.this.add(s);
        }
    }
    
    // GETTERS

    /**
     * returns subtable object for given tableName, null if none exists
     *
     * @param tableName -- SubTableSchema of table whose subtables are to be
     * returned
     * @return subtable object for given tableName
     */
    public SubTableSchema getSubTableSchema(String tableName) {
        for (SubTableSchema s : subtables) {
            if (s.getName().equals(tableName)) {
                return s;
            }
        }
        return null;
    }

    /**
     * returns subtable object for given tableSchema ts, null if none exists
     *
     * @param ts -- TableSchema object for which subtable declaration is to be
     * returned
     * @return subtable object for given tableSchema ts
     */
    public SubTableSchema getSubTableSchema(TableSchema ts) {
        return getSubTableSchema(ts.getName());
    }

    /**
     * returns subtable object for given tableName; throws Error if no such
     * tableName exists
     *
     * @param tableName -- SubTableSchema of table whose subtables are to be
     * returned
     * @return subtable object for given tableName
     */
    public SubTableSchema getSubTableSchemaEH(String tableName) {
        SubTableSchema result = getSubTableSchema(tableName);
        if (result == null) {
            throw Error.toss(Error.tableNoExist, tableName, name);
        }
        return result;
    }

    /**
     * returns subtable object for given tableSchema ts; throws Error if no such
     * ts exists
     *
     * @param ts -- SubTableSchema of tableSchema whose subtables are to be
     * returned
     * @return subtable object for given tableSchema ts
     */
    public SubTableSchema getSubTableSchemaEH(TableSchema ts) {
        SubTableSchema result = getSubTableSchema(ts);
        if (result == null) {
            throw Error.toss(Error.tableNoExist, ts.getName(), name);
        }
        return result;
    }

    /**
     * return list of SubTableSchemas of this database schema
     *
     * @return list of SubTableSchemas of this database schema
     */
    public List<SubTableSchema> getSubTableSchemas() {
        return subtables;
    }

    /**
     * return the TableSchema of the super/belongsToSchema table of TableSchema
     * s
     *
     * @param s -- input TableSchema
     * @return TableSchema of super/belongsToSchema table
     */
    public TableSchema getSuperTable(TableSchema s) {
        for (SubTableSchema sub : subtables) {
            for (TableSchema ts : sub.getSubTableSchemas()) {
                if (ts.getName().equals(s.getName())) {
                    return sub.getSuper();
                }
            }
        }
        return null;
    }

    /**
     * given name in table, return if true if this table has a super or subtable
     *
     * @param tableName --obvious
     * @return true if table has supertable or subtable
     */
    public boolean isTableInHierarchy(String tableName) {
        // Step 1: is table in this database?
        TableSchema ts = getTableSchema(tableName);
        if (ts == null) {
            return false;
        }

        // Step 2: it is, loop through all database subtable declarations
        for (SubTableSchema sts : getSubTableSchemas()) {
            if (ts.equals(sts.getSuper())) {
                return true;
            }
            for (TableSchema tts : sts.getSubTableSchemas()) {
                if (tts.getName().equals(tableName)) {
                    return true;
                }
            }
        }
        // Step 3: if we get here, the answer is false
        return false;
    }

    /**
     * if the table exists, does it have a superTable?
     *
     * @param tableName name of conjectured supertable
     * @return true if 'this' table has tableName as a supertable
     */
    public boolean doesTableHaveSuperTable(String tableName) {
        // Step 1: is table in this database?
        TableSchema ts = getTableSchema(tableName);
        if (ts == null) {
            return false;
        }

        // Step 2: it is, loop through all database subtable declarations
        for (SubTableSchema sts : getSubTableSchemas()) {
            for (TableSchema tts : sts.getSubTableSchemas()) {
                if (tts.getName().equals(tableName)) {
                    return true;
                }
            }
        }
        // Step 3: if we get here, the answer is false
        return false;
    }

    /**
     * is 'this' DBSchema a subset of a bigger schema? reports error string if
     * so, null otherwise
     *
     * @param bigger Schema of bigger
     * @return string of error report
     */
    public String isSubSchemaOf(Schema bigger) {
        ErrorReport er = new ErrorReport();

        // Step 1:  each table of 'this' must belong to bigger
        for (TableSchema ts : getTableSchemas()) {
            TableSchema bts = bigger.getTableSchema(ts.getName());
            if (bts == null) {
                er.add("Table (%s) is in subschema (%s) but not in superschema (%s)", ts.getName(), getName(), bigger.getName());
                continue;
            }
            // Step 1.1: each field of 'this' must belong to bigger also
            for (Column k : ts.getColumns()) {
                Column bk = bts.getColumn(k.getName());
                if (bk == null || !bk.equals(k)) {
                    er.add("Table.column (%s.%s) is in subschema (%s) but not in superschema (%s)", ts.getName(), k.getName(), getName(), bigger.getName());
                    continue;
                }
            }
        }

        // Step 2: each subtableschema of 'this' must belong to bigger
        for (SubTableSchema sts : getSubTableSchemas()) {
            String supName = sts.getSuper().getName();
            SubTableSchema bsts = bigger.getSubTableSchema(supName);
            if (bsts == null) {
                er.add("Table (%s) in subschema (%s) has subtables but not in superschema (%s)", supName, this.getName(), bigger.getName());
                continue;
            }
            for (TableSchema ts : sts.getSubTableSchemas()) {
                if (!bsts.contains(ts.getName())) {
                    er.add("Table (%s) has subtable (%s) in subschema (%s) but not in superschema (%s)", 
                            supName, ts.getName(), this.getName(), bigger.getName());
                } 
            }
        }

        //  Step 3: report and exit
        return er.toString();
    }

    /**
     * is 'this' DBSchema a subset of a bigger schema? Throws errors if so.
     *
     * @param bigger Schema of bigger Throws error if not identical
     */
    public void isSubSchemaOfEH(Schema bigger) {
        String result = isSubSchemaOf(bigger);
        if (result.equals("")) {
            return;
        }
        System.out.println(result);
        System.out.flush();
        throw Error.toss(Error.errorReport, "");
    }

    /**
     * return the TableSchema with SubTableSchema tableName
     *
     * @param tableName of schema to return
     * @return the TableSchema with SubTableSchema tableName, null if not found
     */
    public TableSchema getTableSchema(String tableName) {
        for (TableSchema ts : tableSchemas) {
            if (ts.getName().equals(tableName)) {
                return ts;
            }
        }
        return null;
    }

    /**
     * return the TableSchema with SubTableSchema tableName with built-in error
     * handling
     *
     * @param tableName of schema to return
     * @return the TableSchema with SubTableSchema tableName, throw Error if not
     * found
     */
    public TableSchema getTableSchemaEH(String tableName) {
        TableSchema ts = getTableSchema(tableName);
        if (ts == null) {
            throw Error.toss(Error.tableNoExist, tableName, name);
        }
        return ts;
    }

    /**
     * return list of TableSchemas of this database schema
     *
     * @return list of TableSchemas of this database schema
     */
    public List<TableSchema> getTableSchemas() {
        return tableSchemas;
    }

    /**
     *
     * @param supTable -- name of super table
     * @param subTable -- name of subtable
     * @return true if supTable gtrequ subTable in inheritance hierarchy
     */
    public boolean isGEQ(String supTable, String subTable) {
        // if we're dealing with generics, we're ok
        if (supTable.equals("*") || subTable.equals("*")) {
            return true;
        }
        TableSchema sup = getTableSchemaEH(supTable);
        TableSchema sub = getTableSchemaEH(subTable);
        return isGEQ(sup, sub);
    }

    /**
     *
     * @param sup -- TableSchema of super table
     * @param sub -- TableSchema of subtable
     * @return true if supTabl gtrequ subTable in inheritance hierarchy
     */
    public boolean isGEQ(TableSchema sup, TableSchema sub) {
        // Step 1: isGEQ(sup,sub) == true if sup==sub
        //         that is they are the same schema
        if (sup == sub) {
            return true;
        }

        // Step 2: get subtableschema for sup
        SubTableSchema sts = getSubTableSchema(sup.getName());

        // Step 3: isGEQ(sup,sub)==false if sts==null
        //         because sup has no subtables
        if (sts == null) {
            return false;
        }

        // Step 4: isGEQ(sup,sub)==true if sts.contains(sub)
        if (sts.contains(sub)) {
            return true;
        }

        // Step 5: all else fails, recursively ask isGEQ(st,sub)
        //         for all immediate subtables st of sup
        for (TableSchema st : sts.getSubTableSchemas()) {
            if (isGEQ(st, sub)) {
                return true;
            }
        }
        return false;
    }

    // PRINT UTILITIES
    /**
     * prints to System.out
     */
    public void print() {
        print(System.out);
    }

    /**
     * print the database schema definition to PrintStream out; typically used
     * for demonstrations as it doesn't check the format of the filename; Use
     * print() or print(String).
     *
     * @param out PrintStream destination
     */
    public void print(PrintStream out) {
        out.print(this.toString());
    }

    /**
     * general print to file method calls print(filename,requiredFileEnding)
     * utility method below
     *
     * @param filename -- where to print schema
     */
    abstract public void print(String filename);

    /**
     * print DBSchema in file with filename
     *
     * @param filename -- name of file in which to print DBSchema
     * @param RequiredFileEnding -- files to be written must have this extension
     */
    protected void print(String filename, String RequiredFileEnding) {
        if (!filename.endsWith(RequiredFileEnding)) {
            throw Error.toss(Error.wrongSchemaNameFormat, filename, RequiredFileEnding);
        }
        try {
            PrintStream ps = new PrintStream(filename);
            String x = toString(); // for debugging
            ps.print(x);
            ps.close();
        } catch (IOException ex) {
            throw Error.toss(Error.ioerror, filename, ex.getMessage());
        }
    }

    protected String SchemaDeclString(String whichToken) {
        String db = whichToken + "(" + name + ",[";
        String comma = "";
        for (TableSchema ts : tableSchemas) {
            db = db + comma + ts.getName();
            comma = ",";
        }
        return db + "]).\n";
    }

    @Override
    public abstract String toString(); // calls utility method below

    protected String toString(String whichToken) {
        String result = String.format("%s\n", SchemaDeclString(whichToken));
        for (TableSchema ts : tableSchemas) {
            result = result + ts.toString();
        }
        if (!subtables.isEmpty()) {
            result = result + "\n";
            for (SubTableSchema st : subtables) {
                result = result + st.toString();
            }
        }
        return result;
    }

    /**
     * add identifier 'id' to all tables that are not subtables in 'this'
     * DBSchema do this prior to calling finish methods
     */
    public void addIdentifiersToAllNonSubtables() {
        for (TableSchema ts : getTableSchemas()) {
            boolean isAsubtable = false;
            for (SubTableSchema sts : getSubTableSchemas()) {
                if (sts.contains(ts)) {
                    isAsubtable = true;
                    break;
                }
            }
            if (!isAsubtable) {
                ts.addFirst("id:str");
            }
        }
    }

    /**
     * you finish a schema where you swear that you propagated all attributes in
     * a table hierarchy.
     *
     * @return converted DBSchema object
     */
    public DBSchema finishedAndIPropagatedAttributesMyself() {
        finished = true;
        DBSchema dbs = new DBSchema(name);
        dbs.copyFrom(this); // finished attribute is copied
        return dbs;
    }

    //******************** EDITING OPERATIONS -- USE AT YOUR OWN RISK!! *********//
    
    /**
     * delete table schema ts
     *
     * @param ts -- table to delete
     * @return same as input
     */
    public TableSchema remove(TableSchema ts) {
        tableSchemas.remove(ts);
        return ts;
    }

    /**
     * delete subtable schema ts
     *
     * @param ts -- subtable to delete
     * @return same as input
     */
    public SubTableSchema remove(SubTableSchema ts) {
        subtables.remove(ts);
        return ts;
    }
}
