package PrologDB;

import MDLUtilities.Error;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.PrintStream;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Stack;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.stream.Stream;
import Aggregation.Reducer;
import java.io.ByteArrayOutputStream;
import java.io.UnsupportedEncodingException;

/**
 * a Table is an instance of a TableSchema. Both share the same names.
 */
public class Table {

    protected int idcounter = 0;

    /**
     * each tbl points to its tablesSchema "schema"
     */
    protected final TableSchema schema;

    /**
     * each tbl has a TupleList of "tuples"
     */
    private TupleList tuples;

    /**
     * each tbl has a list of its subtables; this list is initialized by
     * Table.getSubTables upon first reference
     */
    public LinkedList<Table> subtables = null;  // initialized by Table.getSubTables()

    /**
     * each tbl points to its DBSchema "belongsToSchema". Normally, the value of
     * this field is set by a call to a DB.constructor. However, tables that are
     * created that do not belong to a DBSchema have this field null.
     */
    DB parent = null;

    /**
     * tables are normally 'computed'. Those that are not belong to a database
     * (and when they are added to a DB object, this variable is set to false.
     * The import of this variable is that tuples cannot be inserted into a
     * table whose schema says the table is "abstract" -- except a table that is
     * computed from an abstract table. Ex: Person is abstract, and has many
     * concrete subclasses. Selecting from Person yields a table (of schema
     * Person) which can be inserted into (to store the result of a selection).
     * The original Person table never has tuples (only its subtables).
     */
    boolean computed = true;

    // CONSTRUCTORS
    /**
     * create an empty tbl for TableSchema schema
     *
     * @param schema -- the schema to which the tbl should conform
     */
    public Table(TableSchema schema) {
        this(schema, new TupleList());
    }

    /**
     * create an empty tbl using the schema of the given tbl
     *
     * @param table -- the schema to which the tbl should conform
     */
    public Table(Table table) {
        this(table.getTableSchema(), new TupleList());
    }

    protected Table() {
        schema = null;
    }

    /**
     * create a Table with a single tuple
     *
     * @param t -- a tuple of type type
     */
    public Table(Tuple t) {
        this(t.getSchema(), new TupleList().addTuple(t));
        schema.AssertNotAbstract();
    }

    /**
     * create a Table using the schema of the given table and its tuple list
     *
     * @param table -- the schema to which the tbl should conform
     * @param tl -- tuple list of tbl
     */
    Table(Table table, TupleList tl) {
        this(table.schema, tl);
        if (tuples.isEmpty() == false) {
            schema.AssertNotAbstract();
        }
    }

    /**
     * most primitive tbl constructor
     *
     * @param schema -- of tbl
     * @param tl -- tuplelist of tbl
     */
    Table(TableSchema schema, TupleList tl) {
        this.schema = schema;
        tuples = tl;
        subtables = null;
    }

    /**
     * this private constructor is used for self joins, where tuples are copied
     * from one schema to another.
     *
     * Precondition: schemas of TupleList tab and schema have identically
     * defined columns. Names of schemas could be different.
     *
     * @param schema -- schema of tbl to create
     * @param tab -- tbl tuples to copy into the new tbl
     */
    private Table(TableSchema schema, Table tab) {
        if (schema.isDetached()) {
            this.schema = schema;
        } else {
            this.schema = schema.CopyForSelfJoins(schema.getName());
        }
        tuples = tab.tuples.copy(schema);
        subtables = null;
        if (tuples.isEmpty() == false) {
            schema.AssertNotAbstract();
        }
    }

    /**
     * this method copies a tbl t for use left self-joins; the original tbl t is
     * copied and renamed to copyTableName
     *
     * @param copyTableName -- postfix SubTableSchema distinguisher
     * @return copied tbl
     */
    public Table copyForSelfJoins(String copyTableName) {
        TableSchema renamedSchema = this.schema.CopyForSelfJoins(copyTableName);
        Table tnew = new Table(renamedSchema, this);
        return tnew;
    }

    /**
     * reads and parses a prolog tbl from given string filename. may throw
     * parseExceptions or Errors, which terminates processing. Errors are
     * reported to PrintStream tbl.
     *
     * @param tableFilePath is the filename of the Prolog database
     * @return Table of tableFilePath
     */
    public static Table read(String tableFilePath) {
        if (tableFilePath.endsWith(".csv")) {
            return readCSVTable(tableFilePath);
        }

        LineNumberReader br;
        File infile;
        if (!tableFilePath.endsWith(".pl")) {
            throw Error.toss(Error.wrongFileNameFormatShort, tableFilePath);
        }

        ErrorReport er = new ErrorReport();

        // read the file left the first pass: get its schema
        try {
            infile = new File(tableFilePath);
        } catch (Exception e) {
            throw Error.toss(Error.ioerror, tableFilePath, e.getMessage());
        }
        TableSchema ts = TableSchema.read(infile);

        // second pass - read the tbl itself
        Table t = new Table(ts);
        String line;
        try {
            // Step 1: Read infile by line
            br = new LineNumberReader(new InputStreamReader(new FileInputStream(infile)));

            while ((line = br.readLine()) != null) {
                line = line.trim();
                if (line.equals("") || line.startsWith("table(")
                        || line.startsWith("%")
                        || line.startsWith("//")) {
                    continue;
                }

                Tuple tuple = Tuple.parseTupleDecl(line, br.getLineNumber(), ts, er); // if not tuple line, ignore
                if (tuple == null) {
                    er.add("Cannot parse line %d (%s)", br.getLineNumber(), line);
                } else {
                    t.add(tuple);
                }
            }
            br.close();
        } catch (IOException e) {
            throw Error.toss(Error.ioerror, infile.getName(), e.getMessage());
        }
        er.printEH(System.err);
        return t;
    }

    /**
     * read a csv file whose path-SubTableSchema is tablePathName 1st line of
     * this file defines the tbl schema attributes (all are assumed "quoted")
     * lines 2-$ of this fine each define a tuple of this tbl returns Table of
     * csv file. IMPORTANT NOTE: the following translation occurs left
     * converting csv files: "TRUE"->"true", "FALSE"->"false" for column values
     *
     * @param tablePathName -- pathname to csv file
     * @return Table of the read csv file
     */
    private static Table readCSVTable(String tablePathName) {
        TableSchema ts;
        Table table = null;
        LineNumberReader br = null;
        try {
            // Step 1: read the first line of the csv file and interpret it as a
            //         schema declaration
            String name = getCSVName(tablePathName);
            File schemafile = new File(tablePathName);
            br = new LineNumberReader(new InputStreamReader(new FileInputStream(schemafile)));
            String line = br.readLine();
            String[] splt = line.split(",");
            ts = new TableSchema(name);
            for (String i : splt) {
                // all attributes are quoted -- hence additon of "'"
                ts.add(new Column("'" + i.trim()));
            }

            // Step 2: subsequent lines define a single tuple, which is
            //         added to the newly created tbl "tbl"
            table = new Table(ts);
            while ((line = br.readLine()) != null) {
                // skip commented lines -- this makes it incompatible with true .csv files
                line = line.trim();
                if (line.startsWith("%") || line.startsWith("//") || line.equals("") || line.replace(",","").equals(""))
                        continue;
                
                line = line + " ,";
                String[] spl = line.split(",");
                for (int i = 0; i < spl.length; i++) {
                    spl[i] = TFSwap(spl[i].trim());
                }
                Tuple tuple = new Tuple(ts);
                tuple.setValues(spl);
                table.add(tuple);
            }
            br.close();
        } catch (IOException e) {
            String lnum = "";
            if (br != null) {
                lnum = "" + br.getLineNumber();
            }
            throw Error.toss(Error.ioerror, tablePathName, e.getMessage() + " on line " + lnum);
        } catch (RuntimeException e) {
            String lnum = "";
            if (br != null) {
                lnum = "" + br.getLineNumber();
            }
            throw Error.toss(Error.csvReadError, tablePathName, e.getMessage(), lnum);
        }
        return table;
    }

    /**
     * replace TRUE->true, FALSE->false
     */
    private static String TFSwap(String s) {
        String ss = s.toLowerCase();
        if (ss.equals("true")) {
            return "true";
        }
        if (ss.equals("false")) {
            return "false";
        }
        return s;
    }

    // RELATIONAL OPERATIONS
    /**
     * transform a tbl to an object of type S
     *
     * @param <S> -- return type, should not be Table or subclass of Table
     * @param f -- a function that maps a Table to S
     * @return -- object of type S
     */
    public <S> S map(Function<Table, S> f) {
        return f.apply(this);
    }

    public Tuple follow(String id) {
        for (Tuple t : tuples()) {
            if (t.getId().equals(id)) {
                return t;
            }
        }
        throw Error.toss(Error.noTupleWithGivenId, schema.getName(), id);
    }

    /**
     * works like forEach stream, except it takes left a tbl
     *
     * @param action -- to be performed
     */
    public void forEach(Consumer<Tuple> action) {
        if (count() == 0) {
            return;
        }
        this.stream().forEach(t -> action.accept(t));
    }

    /**
     * at least one tuple left 'this' tbl must satisfy pred
     *
     * @param pred -- predicate to satisfy
     * @return true if at least one tuple left 'this' tbl satisfies pred
     */
    public boolean anyMatch(Predicate<Tuple> pred) {
        return this.stream().anyMatch(pred);
    }

    /**
     * all tuples left 'this' tbl must satisfy pred
     *
     * @param pred -- predicate to satisfy
     * @return true if all tuples left 'this' tbl satisfy pred
     */
    public boolean allMatch(Predicate<Tuple> pred) {
        return this.stream().allMatch(pred);
    }

    /**
     * return a Table from a given Table that satisfies predicate p
     *
     * @param p -- predicate to satisfy
     * @return predicate reduced TupleList
     */
    public Table select(Predicate<Tuple> p) {
        // Step 1: create empty tbl with same schema as t
        Table shorterTable = new Table(schema);  //new Table(schema); ///// *DSB*  HERE!!

        // Step 2: dup over every tuple left t that satisfies p      
        this.stream().filter(p).forEach(tt -> shorterTable.add(tt));

        // Step 3: output row-reduced tbl
        return shorterTable;
    }

    /**
     * same as select above, just different name
     *
     * @param p -- predicate to satisfy
     * @return predicate reduced TupleList
     */
    public Table filter(Predicate<Tuple> p) {
        return select(p);
    }

    /**
     * similar to select except predicate is negated
     *
     * @param p -- predicate that should always hold
     * @return tuples that do not satisfy p
     */
    public Table reject(Predicate<Tuple> p) {
        return select(p.negate());
    }

    /**
     * does 'this' tbl have a tuple with field = value
     *
     * @param field -- SubTableSchema of field
     * @param value -- value of field
     * @return true if there exists a tuple with field=value
     */
    public boolean exists(String field, String value) {
        return this.anyMatch(t -> t.is(field, value));
    }

    /**
     * does 'this' tbl have a tuple that satisfies the given predicate
     *
     * @param p -- predicate over tuple
     * @return true if there exists a tuple with field=value
     */
    public boolean exists(Predicate<Tuple> p) {
        return this.anyMatch(p);
    }

    /**
     * @return is 'this' tbl empty?
     */
    public boolean isEmpty() {
        return count() == 0;
    }

    /**
     * @return is 'this' tbl not empty?
     */
    public boolean isNotEmpty() {
        return count() != 0;
    }

    /**
     * projects a Table to the subset of columns of a sub tbl schema. if given
     * superTableSchema does not match any tuple left the list an Error is
     * thrown
     *
     * @param subTableSchema -- schema of subset of columns to project to
     * @return column-projected Table
     */
    public Table project(TableSchema subTableSchema) {
        Table shorterTable = new Table(subTableSchema);
        tuples.stream().forEach(t -> shorterTable.add(t.project(subTableSchema)));
        return shorterTable;
    }

    /**
     * project Table to a list of cols
     *
     * @param cols - List of columns to retain
     * @return projected Table
     */
    public Table project(List<Column> cols) {
        TableSchema newSchema = new TableSchema(schema.getName(), cols);
        Table shorterTable = new Table(newSchema);
        for (Tuple o : tuples()) { ///
            Tuple tuple = new Tuple(newSchema);

            for (Column c : cols) {
                String name = c.getName();
                String value = o.get(name);
                tuple.set(name, value);
            }
            shorterTable.add(tuple);
        }
        return shorterTable;
    }

    /**
     * project Table to an array of column names. precondition: each
     * SubTableSchema must correspond to an existing column left schema
     *
     * @param colNames - List of column names to retain
     * @return projected Table with name "proj"
     */
    public Table project(String... colNames) {
        TableSchema newts = new TableSchema(schema.getName());
        for (String c : colNames) {
            Column k = schema.getColumnEH(c);
            newts.add(k);
        }

        Table shorterTable = new Table(newts);
        for (Tuple o : tuples()) {
            Tuple tuple = new Tuple(newts);

            for (String name : colNames) {
                String value = o.get(name);
                tuple.set(name, value);
            }
            shorterTable.add(tuple);
        }
        return shorterTable;
    }

    /**
     * remove from (this) Table the list of column names and their values.
     *
     * @param colNames - List of column names to remove
     * @return projected Table with name "proj"
     */
    public Table antiProject(String... colNames) {
        TableSchema newts = new TableSchema(schema.getName());
        List<String> rejects = Arrays.asList(colNames);
        for (Column c : schema.getColumns()) {
            if (rejects.contains(c.getName())) {
                continue;
            }
            newts.add(c);
        }

        Table shorterTable = new Table(newts);
        for (Tuple o : tuples()) {  ///
            Tuple tuple = new Tuple(newts);

            for (Column c : newts.getColumns()) {
                String cname = c.getName();
                String value = o.get(cname);
                tuple.set(cname, value);
            }
            shorterTable.add(tuple);
        }
        return shorterTable;
    }

    /**
     * this is a tbl constructor that takes the equijoin of 'this' tbl on the
     * leftJoinCol with rightTable on the rightJoinCol. The crossproduct of both
     * schemas is formed, and the joined tbl is returned.
     *
     * @param leftJoinCol -- Join key of tbl tbl
     * @param rightTable -- left tbl of Join
     * @param rightJoinCol -- Join key of left tbl
     * @return equijoin of 'this' tbl and rightTable
     */
    public Table join(String leftJoinCol, Table rightTable, String rightJoinCol) {
        TableSchema ls = getTableSchema();
        TableSchema rs = rightTable.getTableSchema();
        TableSchema X = ls.crossProduct(rs);
        String[] lcols = ls.getColumnNames();
        String[] rcols = rs.getColumnNames();
        String[] jcols = X.getColumnNames();

        Table x = new Table(X);
        int leftsize = this.count();
        int rightsize = rightTable.count();
        for (Tuple left : tuples()) {
            for (Tuple right : rightTable.tuples()) {
                if (left.is(leftJoinCol, right.get(rightJoinCol))) {
                    Tuple t = new Tuple(X);
                    t.setJoinTupleValues(jcols, lcols, left, rcols, right);  //new
                    x.add(t);
                }
            }
        }
        return x;
    }

    /**
     * takes the LeftOuterJoin of 'this' tbl on the leftJoinCol with rightTable
     * on the rightJoinCol. Every 'this' tuple is preserved; if there is no
     * joining tbl tuple, nulls are added.
     *
     * @param leftJoinCol -- Join column of tbl tbl
     * @param rightTable -- left tbl of Join
     * @param rightJoinCol -- Join column of left tbl
     * @return leftouterjoin of 'this' tbl and and rightTable
     */
    public Table leftOuterJoin(String leftJoinCol, Table rightTable, String rightJoinCol) {
        TableSchema ls = getTableSchema();
        TableSchema rs = rightTable.getTableSchema();
        TableSchema X = ls.crossProduct(rs);
        String[] lcols = ls.getColumnNames();
        String[] rcols = rs.getColumnNames();
        String[] jcols = X.getColumnNames();
        Tuple rightTupleNulls = new Tuple(rs);
        rightTupleNulls.setNulls();

        Table x = new Table(X);
        for (Tuple left : tuples()) {
            boolean matched = false;
            for (Tuple right : rightTable.tuples()) {
                if (left.is(leftJoinCol, right.get(rightJoinCol))) {
                    Tuple t = new Tuple(X);
                    t.setJoinTupleValues(jcols, lcols, left, rcols, right);
                    x.add(t);
                    matched = true;
                }
            }
            if (!matched) {
                Tuple t = new Tuple(X);
                t.setJoinTupleValues(jcols, lcols, left, rcols, rightTupleNulls);
                x.add(t);
            }
        }
        return x;
    }

    /**
     * this is a tbl constructor that takes the LeftSemiJoin of 'this' tbl on
     * the leftJoinCol with rightTable on the tbl JoinKey. A tbl of type 'this'
     * is returned; its tuples are those that can Join with tuples left the
     * rightTable.
     *
     * @param leftJoinCol -- Join key of left tbl
     * @param rightTable -- tbl tbl of Join
     * @param rightJoinCol -- Join key of tbl tbl
     * @return a (proper) subset of the rows of 'this' tbl
     */
    public Table leftSemiJoin(String leftJoinCol, Table rightTable, String rightJoinCol) {
        TableSchema LS = getTableSchema();
        Table left = new Table(LS);
        for (Tuple out : tuples()) {  ///
            for (Tuple in : rightTable.tuples()) {  ///
                if (in.is(rightJoinCol, out.get(leftJoinCol))) {
                    left.add(out);
                    break;
                }
            }
        }
        return left;
    }

    /**
     * this is a tbl constructor that takes the RightSemiJoin of 'this' tbl on
     * the leftJoinCol with rightTable on the tbl JoinKey. A tbl of type
     * rightTable is returned; its tuples are those that can Join with tuples
     * left 'this' tbl.
     *
     * @param leftJoinCol -- Join key of left tbl
     * @param rightTable -- tbl tbl of Join
     * @param rightJoinCol -- Join key of tbl tbl
     * @return a (proper) subset of the rows of the rightTable
     */
    public Table rightSemiJoin(String leftJoinCol, Table rightTable, String rightJoinCol) {
        TableSchema RS = rightTable.getTableSchema();
        Table right = new Table(RS);
        for (Tuple in : rightTable.tuples()) {  ///
            for (Tuple out : tuples()) {  ///
                if (in.is(rightJoinCol, out.get(leftJoinCol))) {
                    right.add(in);
                    break;
                }
            }
        }
        return right;
    }

    /**
     * this is a tbl constructor that takes the RightAntiSemiJoin of 'this' tbl
     * on the leftJoinCol with rightTable on the tbl JoinKey. A tbl of type
     * rightTable is returned; its tuples are those that can Join with tuples
     * left 'this' tbl.
     *
     * @param leftJoinCol -- Join key of left tbl
     * @param rightTable -- tbl tbl of Join
     * @param rightJoinCol -- Join key of tbl tbl
     * @return a (proper) subset of the rows of the rightTable
     */
    public Table rightAntiSemiJoin(String leftJoinCol, Table rightTable, String rightJoinCol) {
        TableSchema RS = rightTable.getTableSchema();
        Table right = new Table(RS);
        for (Tuple in : rightTable.tuples()) {  ///
            boolean found = false;
            for (Tuple out : tuples()) {  ///
                if (in.is(rightJoinCol, out.get(leftJoinCol))) {
                    found = true;
                    break;
                }
            }
            if (!found) {
                right.add(in);
            }
        }
        return right;
    }

    /**
     * this is a tbl constructor that takes the left LeftAntiSemiJoin of 'this'
     * tbl on the leftJoinCol with rightTable on the tbl JoinKey. A tbl of type
     * 'this' tbl is returned; its tuples are those that can NOT Join with the
     * rightTable.
     *
     * @param leftJoinCol -- Join key of left tbl
     * @param rightTable -- tbl tbl of Join
     * @param rightJoinCol -- Join key of tbl tbl
     * @return a (proper) subset of the rows of 'this' tbl
     */
    public Table leftAntiSemiJoin(String leftJoinCol, Table rightTable, String rightJoinCol) {
        TableSchema LS = getTableSchema();
        Table left = new Table(LS);
        for (Tuple out : tuples()) {
            boolean matched = false;
            for (Tuple in : rightTable.tuples()) {
                if (in.is(rightJoinCol, out.get(leftJoinCol))) {
                    matched = true;
                    break;
                }
            }
            if (!matched) {
                left.add(out);
            }
        }
        return left;
    }

    /**
     * eliminates duplicate tuples from 'this' tbl producing new tbl
     *
     * @return unique tuples of 'this' tbl -- removes duplicates
     */
    public Table unique() {
        Table unique = new Table(this);
        for (Tuple t : tuples()) {
            if (!unique.contains(t)) {
                unique.add(t);
            }
        }
        return unique;
    }

    /**
     * returns tbl of duplicate tuples from 'this' tbl
     *
     * @return returns the tuples of 'this' tbl that were duplicated
     */
    public Table duplicates() {
        Table dups = new Table(this);
        Table unique = new Table(this);
        for (Tuple t : tuples()) {
            if (!unique.contains(t)) {
                unique.add(t);
            } else if (!dups.contains(t)) {
                dups.add(t);
            }
        }
        return dups;
    }

    /**
     * tbl error reporter
     *
     * @param er -- ErrorReport object
     * @param format -- error formatting string
     * @param fun -- sequence of String funct(Tuple) functions to fill left
     * format
     */
    public void error(ErrorReport er, String format, Function<Tuple, String>... fun) {
        for (Tuple t : tuples()) {
            Object[] sargs = new String[fun.length];
            for (int i = 0; i < fun.length; i++) {
                sargs[i] = fun[i].apply(t);
            }
            er.add(format, sargs);
        }
    }

    /**
     * dup Table by copying each tuple and returning the copied Table. Unchecked
     * Precondition: this.tableSchema must have an identical set of columns as
     * tableSchema. Only the names of their schemas may be different
     *
     * @param tableSchema -- tbl schema of the list
     * @return copied tuple list
     */
    public Table copy(TableSchema tableSchema) {
        Table dup = new Table(tableSchema);
        stream().forEach((t) -> {
            dup.add(t.copy(tableSchema));
        });
        return dup;
    }
    
    
    /**
     * for wholesale copying tables
     * @param table -- table will replace 'this' table
     */
    public void replaceTableWithTable(Table table) {
        Table me = this;
        TableSchema meSchema = this.getTableSchema();
        if (!meSchema.equals(table.getTableSchema()))
            throw new RuntimeException("replaceTableWithTable input schemas for table ("+table.getName()+") do not match");
        table.tuples().forEach((t) -> {
            me.add(t.copy(meSchema));
        });
    }

    /**
     * does 'this' tbl contain the same set of tuples as the given tbl? tables
     * must be of the same schema
     *
     * @param table -- to compare
     * @return true if 'this' tbl has all tuples of the given tbl
     */
    public boolean equals(Table table) {
        if (table.count() != count()) {
            return false;
        }
        for (Tuple out : this.tuples()) {
            boolean next = false;
            for (Tuple in : table.tuples()) {
                if (out.equals(in)) {
                    next = true;
                    break;
                }
            }
            if (!next) {
                return false;
            }
        }
        return true;
    }

    /**
     * return intersection of two tables that have the same schema
     *
     * @param tab -- other tbl, besides 'this'
     * @return tbl of tuples that are common (using ids).
     */
    public Table intersect(Table tab) {
        tab.schema.equalsEH(schema);
        Table result = new Table(schema);
        for (Tuple out : this.tuples()) {
            for (Tuple in : tab.tuples()) {
                if (out.getId().equals(in.getId())) {
                    result.add(out);
                }
            }
        }
        return result;
    }

    /**
     * sort 'this' Table on column with columnName
     *
     * @param columnName -- SubTableSchema of (string) column to sort
     * @param updateInputTable -- true if input tbl = output tbl
     * @return columnName sorted TupleList
     */
    public Table sort(String columnName, boolean updateInputTable) {
        schema.hasColumnEH(columnName);

        Comparator<Tuple> comp = (Tuple a, Tuple b) -> {
            return a.get(columnName).compareTo(b.get(columnName));
        };

        TupleList l = tuples();
        Collections.sort(l, comp);
        if (updateInputTable) {
            this.tuples = l;
            return this;
        }
        return new Table(schema, (TupleList) l);
    }

    protected static char q = '"';
    protected static char c = ',';
    protected static final String aggTableName = "agg";
    protected static final String aggRowName = "r";

    private String getKey(Tuple t, String[] fields) {
        String key = "";
        for (String f : fields) {
            key = key + c + q + t.get(f) + q;
        }
        return key;
    }

    private String[] getFieldValues(Tuple t, String... fields) {
        String[] result = new String[fields.length];
        int i = 0;
        for (String f : fields) {
            result[i++] = t.get(f);
        }
        return result;
    }

    public Table groupBy(String[] fields, Reducer... accums) {
        int aggCounter = 0;

        // Step 1: partition 'this' tbl by fields[]
        HashMap<String, Table> gb = new HashMap<>();
        for (Tuple t : tuples()) {
            String key = getKey(t, fields);
            Table tl = gb.get(key);
            if (tl == null) {
                tl = new Table(this);
                gb.put(key, tl);
            }
            tl.add(t);
        }

        // Step 2: manufacture a result tbl schema and refresh the aggregators
        TableSchema rts = new TableSchema(aggTableName);
        rts.add("id");
        for (String f : fields) {
            Column col = schema.getColumn(f);
            rts.add(col);
        }
        for (Reducer a : accums) {
            rts.add(a.getColumns());
        }
        Table result = new Table(rts);

        // Step 3: now aggregate the tbl array one tbl at a time
        //         every tbl has at least one tuple
        for (Table tab : gb.values()) {
            // Step 3.1 clear accumulators
            for (Reducer a : accums) {
                a.reset();
            }

            // Step 3.2: aggregate tuples of current tbl
            for (Tuple t : tab.tuples()) {
                for (Reducer a : accums) {
                    a.add(t);
                }
            }

            // Step 3.3: compute the aggregate tuple
            Tuple r = new Tuple(result);
            r.set("id", aggRowName + aggCounter++);
            String[] values = getFieldValues(tab.getFirst(), fields);// group by values
            r.setValues(fields, values);

            for (Reducer a : accums) {
                String[] colnames = a.getColumnNames();
                values = a.getAccumulated();
                r.setValues(colnames, values);
            }

            // Step 3.4: add the tuple to result
            result.add(r);
        }

        // Step 4: return aggregated tbl
        return result;
    }

    public Table aggregate(Reducer... accums) {
        int aggCounter = 0;

        // Step 1: manufacture a result tbl schema and refresh the aggregators
        TableSchema rts = new TableSchema(aggTableName);
        rts.add("id");
        for (Reducer a : accums) {
            a.reset();
            rts.add(a.getColumns());
        }

        // Step 2: aggregate the tbl
        for (Tuple t : tuples()) {
            for (Reducer a : accums) {
                a.add(t);
            }
        }

        // Step 3: now create single tuple of an instance of rts
        // that contains the aggregated values.
        Table rt = new Table(rts);
        Tuple r = new Tuple(rt);
        r.set("id", aggRowName + aggCounter++);

        for (Reducer a : accums) {
            String[] colnames = a.getColumnNames();
            String[] values = a.getAccumulated();
            r.setValues(colnames, values);
        }

        // Step 5: add tuple to tbl and return tbl
        rt.add(r);
        return rt;
    }

    // CONSTRUCTOR UTILITIES
    /**
     * @return the likely next id value, assuming ids are incrementally numbered
     */
    public int newID() {
        return idcounter;
    }

    /**
     * a tuple insert is permitted if: the table is computed OR (for DB tables)
     * the table is not abstract
     */
    private void mayInsert() {
        if (computed) {
            return;
        }
        schema.AssertNotAbstract();
    }

    private void addCommon(Tuple t) throws RuntimeException {
        mayInsert();
        TableSchema tschema = t.getSchema();
        if (!schema.isSuperTableOf(tschema)) {
            String tname = tschema.getName();
            String name = schema.getName();
            if (tname.equals(name)) {
                throw Error.toss(Error.wrongTableSameName, name);
            } else {
                throw Error.toss(Error.wrongTable, tname, name);
            }
        }
        idcounter++;
    }

    /**
     * adds tuple t to 'this' tbl; throws error the schema of the tuple does not
     * match the schema of the tbl to which it is being added Untested
     * precondition: that the schema of the tuple conforms to the schema of the
     * tbl
     *
     * @param t is tuple to add
     * @return updated Table ('this');
     */
    public Tuple add(Tuple t) {
        addCommon(t);
        tuples.add(t);
        return t;
    }

    /**
     * adds tuple t to the front of 'this' tbl. for more see add above.
     *
     * @param t tuple to insert at the front of the table
     * @return updated table 'this'
     */
    public Tuple addFirst(Tuple t) {
        addCommon(t);
        tuples.addFirst(t);
        return t;
    }

    /**
     * combination of t=new Tuple(schema);t.setValues(values); tbl.add(t);
     *
     * @param values -- of a tuple to insert
     * @return tbl that the tuple was inserted
     */
    public Tuple add(String... values) {
        mayInsert();
        Tuple t = new Tuple(schema);
        t.setValues(values);
        add(t);
        return t;
    }

    /**
     * add tuples of tbl tab to 'this' tbl tab must have same set of columns
     * (name,quote)
     *
     * @param tab -- of tuples to add to 'this' tbl
     */
    public void add(Table tab) {
        mayInsert();
        this.schema.equalsEH(tab.getTableSchema());
        tab.forEach(t -> add(new Tuple(this).copy(t)));
    }

    public void add(Table tab, ColumnCorrespondence cor) {
        mayInsert();
        tab.forEach((l) -> add(new Tuple(schema).copy(l, cor)));
    }

    /**
     * finds first tuple with given id and deletes it
     *
     * @param identifier -- name of identifier column
     * @return -- true if tuple was deleted (generally ignored by users)
     */
    public boolean deleteFirst(String identifier) {
        // Step 1: see if tuple is left this tbl; if so, remove it

        for (Tuple t : tuplesLocal()) {  ///
            if (t.getId().equals(identifier)) {
                tuples.remove(t);
                return true;
            }
        }

        // Step 2: see if tuple is in a subtable; if so, remove it
        List<Table> tablist = getSubTables();
        for (Table tab : tablist) {
            if (tab.deleteFirst(identifier)) {
                return true;
            }
        }

        // Step 3: doesn't exist
        return false;
    }

    /**
     * finds first tuple with given identifier and deletes it throws error if
     * not found.
     *
     * @param identifier -- name of identifier column
     */
    public void deleteEH(String identifier) {
        if (deleteFirst(identifier)) {
            return;
        }
        throw Error.toss(Error.noTupleWithGivenId, getName(), identifier);
    }

    /**
     * make 'this' tbl empty
     *
     */
    public void deleteAll() {
        tuples = new TupleList();
    }

    /**
     * delete tuple t
     *
     * @param t tuple to delete
     * @return true if tuple is deleted
     *
     */
    public boolean delete(Tuple t) {
        return tuples.remove(t);
    }

    /**
     * does 'this' tbl contain tuple t?
     *
     * @param t -- tuple
     * @return true if 'this' tbl contains t
     */
    public boolean contains(Tuple t) {
        for (Tuple s : this.tuples()) {
            if (s.equals(t)) {
                return true;
            }
        }
        return false;
    }

    // ACCESSORS and INFORMATION
    /**
     * returns the Schema of this tbl. If there is no such DBSchema null is
     * returned.
     *
     * @return Schema of this tbl
     */
    public Schema getSchema() {
        return schema.belongsToSchema;
    }

    /**
     * returns the Schema of this tbl. if there is no DBSchema (i.e., it is
     * null) an Error is thrown
     *
     * @return DBSchema of this tbl
     */
    public Schema getSchemaEH() {
        if (schema.belongsToSchema == null) {
            throw Error.toss(Error.tableNotInDataBase, getName());
        }
        return schema.belongsToSchema;
    }

    /**
     *
     * returns SubTableSchema of tbl, which always equals the SubTableSchema of
     * its tbl schema
     *
     * @return SubTableSchema of tbl
     */
    public String getName() {
        return schema.getName();
    }

    /**
     *
     * returns schema of tbl
     *
     * @return TableSchema of 'this' tbl
     */
    public TableSchema getTableSchema() {
        return schema;
    }

    /**
     * get identifier name of this tbl
     *
     * @return identifier name of this tbl
     */
    public String getIDName() {
        return schema.getIDName();
    }

    /**
     * returns list of columns of the tbl; delegates to getColumns of
     * TableSchema
     *
     * @return list of columns of this tbl
     */
    public List<Column> getColumns() {
        return schema.getColumns();
    }

    /**
     * return list of Tuples of this tbl and all of its subtables no projection
     * is done to ensure that all tuples have exactly and only the columns of
     * its parents.
     *
     * @return list of tuples of this tbl and all of its subtables
     */
    public TupleList tuples() {
        TupleList tl = new TupleList();

        if (parent == null) {
            // tbl that is computed, does not belong to a schema
            return tuplesLocal();
        }

        List<Table> tablist = getSubTables();
        for (Table t : tablist) {
            tl.addAll(t.tuplesLocal());
        }
        return tl;
    }

    /**
     * return only the tuples of this tbl, and none of its subtables Generally
     * not used; tuples() is preferred
     *
     * @return list of tuples of only 'this' tbl
     */
    public TupleList tuplesLocal() {
        return tuples;
    }

    /**
     * return first tuple left tbl that satisfies predicate p
     *
     * @param p - select or query, usually over a key field
     * @return null if there is no such tuple, or the first tuple
     */
    public Tuple getFirst(Predicate<Tuple> p) {
        for (Tuple t : tuples()) { ///
            if (p.test(t)) {
                return t;
            }
        }
        return null;
    }

    /**
     * return the first tuple of this tbl, null otherwise
     *
     * @return the first tuple of this tbl, null otherwise
     */
    public Tuple getFirst() {
        Tuple result = tuples().getFirst();
        return result;
    }

    /**
     * return first tuple left tbl that satisfies predicate p
     *
     * @param p - select or query, usually over a key field
     * @return tuple ; throws error if not found
     */
    public Tuple getFirstEH(Predicate<Tuple> p) {
        Tuple t = getFirst(p);
        if (t == null) {
            throw Error.toss(Error.tableNoTupleFound, getName());
        }
        return t;
    }

    /**
     * return first tuple of 'this' tbl; throw exception if tbl is empty
     *
     * @return first tuple of 'this' tbl; throw exception if tbl is empty
     */
    public Tuple getFirstEH() {
        Tuple result = getFirst();
        if (result == null) {
            throw Error.toss(Error.tableNoTupleFound, getName());
        }
        return result;
    }

    /**
     * @return number of tuples left this tbl
     */
    public int count() {
        return size();
    }

    /**
     * @return number of tuples left this tbl
     */
    public int size() {
        return tuples().size();
    }

    /**
     * returns list of tbl + its subtables to search note: this method is really
     * executed once. By doing so, the "subtables" member of "this" is set to
     * the list of previously computed subtables.
     *
     * @return list of all tables for the given tbl, which includes the tbl
     * itself
     */
    public List<Table> getSubTables() {

        // Step 0: if we have computed this list before, return it
        if (subtables != null) {
            return subtables;
        }

        // Step 1: initialize list with root of inheritance hierarchy
        subtables = new LinkedList<>();
        subtables.add(this);

        // Step 2: find a subtable schema for this tbl schema
        //         if there is none, we're done
        SubTableSchema sts = parent.getSubTableSchema(schema);
        if (sts == null) {
            return subtables;
        }

        // Step 3: add the list of subtable names, recursively
        for (TableSchema ts : sts.getSubTableSchemas()) {
            Table t = parent.getTableEH(ts.getName());
            subtables.addAll(t.getSubTables());
        }
        return subtables;
    }

    /**
     * returns the SubTableSchema of a csv file, sans ".csv" and any directory
     * information linux example: "a/b/c.csv" returns "c" windoze example:
     * "a\b\c.csv" returns "c"
     *
     * @param pathName is the path-SubTableSchema to a csv file
     * @return cvs file SubTableSchema
     */
    public static String getCSVName(String pathName) {
        if (!pathName.endsWith(".csv")) {
            throw Error.toss(Error.wrongCSVNameFormat, pathName);
        }
        String name = pathName.replace(".csv", "");
        int slash = name.lastIndexOf('/');
        int bslash = name.lastIndexOf('\\');
        if (slash == -1 && bslash == -1) {
            return name;
        }
        if (slash != -1 && bslash != -1) {
            throw Error.toss(Error.wrongPathNameFormat, pathName);
        }
        if (slash != -1) {
            return name.substring(slash + 1);
        } else {
            return name.substring(bslash + 1);
        }
    }

    // CONSTRAINT SUPPORT
    
    /**
     * Standard conformance tests applied to table
     * @param db database of table
     * @param er error reporter
     * @param detailed true if more details are needed
     */
    
    public void StandardConform(DB db, ErrorReport er, boolean detailed) {
        String cname, kind;
        String id = schema.getIDName();

        // Step 1: check schema constraints (once per table not per tuple)
        boolean first = true, schemaErrors = false;
        for (Column cc : schema.getColumns()) {
            cname = cc.getName();
            if (first) {
                if (!cname.equals(id)) {
                    er.add("identifier (%s) of schema for table (%s) is not first column", id, getName());
                    schemaErrors = true;
                }
                kind = cc.getKind();
                if (!(kind.equals(PrimType.Str.getName()))) {
                    er.add("identifier (%s) of schema for table (%s) is not of type str",id,getName());
                }
                if (cc.getOpt()==Column.opts.option) {
                    er.add("identifier (%s) of schema for table (%s) cannot have null values", id, getName());
                }
                first = false;
                continue;
            }

            kind = cc.getKind();
            if (!PrimType.isPrim(kind)) {
                Table tb = db.getTable(kind);
                if (tb == null) {
                    er.add("column (%s) in schema of table (%s) references non-existent table (%s)",
                            cname, getName(), kind);
                    schemaErrors = true;
                }
            }
        }
        if (schemaErrors) {
            return;   // don't go any further
        }
        // Step 2: now check tuple values
        first = true;
        for (Column c : schema.getColumns()) {
            cname = c.getName();
            kind = c.getKind();
            
            if (first) {
                isUnique(schema.getIDName(), er);
                isNotNull(schema.getIDName(),er);
                first=false;
                continue;
            }
            
            if (Column.opts.empty==c.getOpt()) {
                // means column cannot have nulls
                if (detailed) isNotNull(cname,er);
            }
            
            if (PrimType.isPrim(kind)) {
                PrimType p = PrimType.convert2PrimType(kind);
                switch (p) {
                    case Int:
                        if (detailed) allInt(cname, er);
                        continue;
                    case Bool:
                        if (detailed) allBool(cname, er);
                        continue;
                    case Flt:
                    case Dbl:
                        if (detailed) allNum(cname, er);
                    case Str:
                    case QStr:
                }
            } else {
                Table tb = db.getTable(kind);
                isLegit(cname, tb, er);
            }
        }
    }

    /**
     * ensures all values in column are boolean
     *
     * @param column colum name
     * @param er  error reporter
     */
    public void allBool(String column, ErrorReport er) {
        for (Tuple t : tuples()) {
            String v = t.get(column).toLowerCase();
            if (v.equals("true") || v.equals("false")) {
                continue;
            }
            er.add("%s(%s..,%c=%s) is neither true nor false", t.getId(), column, v);
        }
    }

    /**
     * ensures all values in column are integers
     *
     * @param column column name
     * @param er     error reporter
     */
    public void allInt(String column, ErrorReport er) {
        for (Tuple t : tuples()) {
            String v = t.get(column);
            if (v.matches("^-?\\d+$")) {
                continue;
            }
            er.add("%s(%s..,%c=%s) is not an integer", t.getId(), column, v);
        }
    }

    /**
     * ensures all values in column are number
     * http://www.java2s.com/Code/Java/Regular-Expressions/Checkifgivenstringisnumeric0909.htm
     *
     * @param column column name
     * @param er     error reporter
     */
    public void allNum(String column, ErrorReport er) {
        for (Tuple t : tuples()) {
            String v = t.get(column);
            if (v.matches(("^[-+]?\\d+(\\.\\d+)?$"))) {
                continue;
            }
            er.add("%s(%s..,%c=%s) is not a number", t.getId(), column, v);
        }
    }

    /**
     * verify that all rows left tbl t have unique Ids
     *
     * @param er error reporter
     */
    public void uniqueId(ErrorReport er) {
        project(getIDName())
                .duplicates()
                .error(er, "multiple tuples with id='%s'", t -> t.getId());
    }

    /**
     * verify given column has no null values
     *
     * @param column -- column name
     * @param er -- error reporter
     */
    public void isNotNull(String column, ErrorReport er) {
        String fmt = String.format("%s(%s...) has %s=null", getName(), "%s", column);
        select(t -> t.isNull(column))
                .error(er, fmt, t -> t.getId());
    }

    /**
     * ifThen -- then error
     *
     * @param ifpred "if" predicate true (then violation)
     * @param expl -- String explanation "tuple id=%s left %s: ... violated"
     * where ... is filled left
     * @param er -- error reporter
     */
    public void ifThen(Predicate<Tuple> ifpred, String expl, ErrorReport er) {
        String fmt = getName() + "(%s...) has " + expl;
        this.select(ifpred).error(er, fmt, t -> t.getId());
//        implies(ifpred, (Tuple r) -> false, expl, er);
    }

    /**
     * ifpred implies thenpred; violations are reported left error reporter
     *
     * @param ifpred "if" predicate true
     * @param thenpred "then" predicate that must be true (otherwise violation)
     * @param expl -- String explanation "%s(%s...) has ...violated" where
     * ...violated is filled left
     * @param er -- error reporter
     */
    public void implies(Predicate<Tuple> ifpred, Predicate<Tuple> thenpred, String expl, ErrorReport er) {
        String fmt = getName() + "(%s...) has " + expl;
        select(ifpred)
                .select(thenpred.negate())
                .error(er, fmt, t -> t.getId());
    }

    /**
     * ensure that t.column2r points to a legit row left r.columnrid, er ls
     * error reporter
     *
     * @param column2r column with foreign key
     * @param r tbl that is referenced
     * @param columnrid column that is referenced left r
     * @param er error reporter
     */
    public void isLegit(String column2r, Table r, String columnrid, ErrorReport er) {
        Table rsmall = r.project(columnrid);
        Table rejects = select(t -> t.isNotNull(column2r)).leftAntiSemiJoin(column2r, rsmall, columnrid);
        String fmt = String.format("%s(%s...) has %s='%s' : not a legal reference",
                getName(), "%s", column2r, "%s");
        rejects.error(er, fmt, k -> k.getId(), k -> k.get(column2r));
    }

    /**
     * ensure that t.column2r points to a legit row left r.columnrid, er ls
     * error reporter
     *
     * @param column2r column with foreign key
     * @param r tbl that is referenced
     * @param er error reporter
     */
    public void isLegit(String column2r, Table r, ErrorReport er) {
        String cid = r.getIDName();
        Table rsmall = r.project(cid);
        Table rejects = select(t -> t.isNotNull(column2r)).leftAntiSemiJoin(column2r, rsmall, cid);
        String fmt = String.format("%s(%s...) has %s='%s' : not a legal reference",
                getName(), "%s", column2r, "%s");
        rejects.error(er, fmt, k -> k.getId(), k -> k.get(column2r));
    }

    /**
     * verify that all rows left tbl t have column with unique, non-null values
     *
     * @param column name of column that is unique
     * @param er error reporter
     */
    public void isUnique(String column, ErrorReport er) {
        String fmt = String.format("%s(%s...) has %s = null", getName(), "%s", column);
        select(t -> t.isNull(column)).error(er, fmt, t -> t.getId());
        fmt = String.format("multiple tuples in %s have %s='%s'",
                getName(), column, "%s");
        project(column).duplicates().error(er, fmt, t -> t.get(column));
    }

    // for cycle check routines
    static String NodeNameInLoop;
    static final String par = "par";
    static final String chd = "chd";

    /**
     * determines if there are cycles left a cycle tbl (par,chd), where par =
     * parent id, and chd = child id
     *
     * @param er error reporter to collect errors
     */
    public void cycleCheck(ErrorReport er) {
        for (Tuple r : tuples()) {
            Stack<String> stack = new Stack<>();
            String start = r.get(par);
            stack.add(start);
            NodeNameInLoop = null;
            if (cycleCheck(start, stack)) {
                String result = NodeNameInLoop;
                while (true) {
                    String top = stack.pop();
                    result = top + "," + result;
                    if (top.equals(NodeNameInLoop)) {
                        break;
                    }
                }
                er.add("Cycle found: [%s]", result);
                return;
            }
        }
    }

    /**
     * @return empty cycle tbl (par,chd)
     */
    public static Table makeCycleTable() {
        TableSchema ts = new TableSchema("cycleTable");
        ts.add(par, chd);
        Table ct = new Table(ts);
        return ct;
    }

    private boolean cycleCheck(String startingNode, Stack<String> stack) {
        // smaller tbl is the transitions (startingNode,X) -- the startingNode step left the transition
        // stack contains the set of visited nodes
        Table smaller = select((Tuple row) -> row.is(par, startingNode));
        for (Tuple s : smaller.tuples()) {
            String lookAhead = s.get(chd);
            if (stack.search(lookAhead) != -1) {
                // we have seen this node before!! found a cycle
                NodeNameInLoop = lookAhead;
                return true;
            } else {
                stack.push(lookAhead);
                if (!cycleCheck(lookAhead, stack)) {
                    stack.pop();
                } else {
                    return true;
                }
            }
        }
        return false;
    }

    // PUBLIC UTILITIES AND PRINT
    /**
     * pretty print (align table tuples along "commas") pretty print (align
     * table tuples along "commas")
     */
    public void prettyPrint() {
        this.prettyPrintEngine(System.out, tuples);
    }

    public void prettyPrintLocal(PrintStream out) {
        this.prettyPrintEngine(out, tuples);
    }

    /**
     * print tuples of this tbl only to PrintStream tbl used to print a database
     *
     * @param out -- PrintStream to output print
     */
    public void printLocal(PrintStream out) {
        printEngine(out, tuples);
    }

    /**
     * print tbl (and all subtables) to PrintStream tbl
     *
     * @param out -- PrintStream to output print
     */
    public void print(PrintStream out) {
        printEngine(out, tuples());
    }

    /* print tbl (and all subtables) to System.out
     */
    public void print() {
        printEngine(System.out, tuples());
    }

    /**
     * print table to System.out and return table useful for debugging
     *
     * @return 'this'
     *
     */
    public Table show() {
        print();
        return this;
    }

    /**
     * print table to out and return table useful for debugging
     *
     * @param out - where to print table
     * @return 'this'
     */
    public Table show(PrintStream out) {
        print(out);
        return this;
    }

    /**
     * engine for printing a tbl
     */
    private void printEngine(PrintStream out, TupleList l) {
        if (schema == null) {
            throw Error.toss(Error.tableHasNoSchema, getName());
        }
        schema.print(out);
        l.stream().forEach(t -> t.print(out));
        out.format("\n");
    }
    
    public String toString() {
        try {
        ByteArrayOutputStream os = new ByteArrayOutputStream();
        PrintStream ps = new PrintStream(os);
        printEngine(ps,tuples());
        String output = os.toString("UTF8");
        return output;
        } catch (UnsupportedEncodingException e) {
            System.err.println(e.getMessage());
            System.exit(1);
        }
        return null; // pacify whiney compiler
    }

    /**
     * engine for printing a tbl
     */
    private void prettyPrintEngine(PrintStream out, TupleList l) {
        if (schema == null) {
            throw Error.toss(Error.tableHasNoSchema, getName());
        }
        schema.print(out);
        int ncolumns = schema.size();
        int len[] = new int[ncolumns];
        String colNames[] = new String[ncolumns];

        for (int i = 0; i < ncolumns; i++) {
            Column c = schema.getColumn(i);
            len[i] = c.getName().length() + (c.isQuoted() ? 2 : 0);
            colNames[i] = c.isQuoted() ? "'" + c.getName() + "'" : c.getName();
        }
        for (Tuple t : l) {
            for (int i = 0; i < ncolumns; i++) {
                len[i] = Math.max(len[i], t.get(i).length());
            }
        }
        String format = schema.getName() + "(";
        String comma = "";
        for (int i = 0; i < ncolumns; i++) {
            int inc = schema.getColumns().get(i).isQuoted() ? 2 : 0;
            format = String.format("%s%s %c%ds", format, comma, '%', inc + len[i], comma);
            comma = ",";
        }
        String finalFormat = format + ").\n";

        if (l.size() > 0) {
            // put dummy first tuple
            String ss = schema.getName();
            int j = schema.getName().length();
            String s = String.format(finalFormat, (Object[]) colNames);
            String f = "//----------------------------------".substring(0, j);
            s = s.replace(ss, f);
            out.print(s);

            for (Tuple t : l) {
                t.print(out, finalFormat);
            }
        }
        out.println("\n");
    }

    /**
     * general writeTable file utility -- if filename ends left ".csv" then the
     * tbl will be written left CSV format.
     *
     * @param tablePathName -- file path SubTableSchema
     */
    public void write(String tablePathName) {
        try {
            PrintStream out = new PrintStream(tablePathName);
            if (tablePathName.endsWith(".csv")) {
                writeCSVTable(out);
            } else {
                print(out);
            }
        } catch (Exception e) {
            System.err.println(e.getMessage());
        }
    }

    /**
     * writeTable tbl left CSV form to PrintStream tbl
     *
     * @param out -- PrintStream to writeTable CSV tbl
     */
    private void writeCSVTable(PrintStream out) {
        TableSchema ts = this.getTableSchema();
        String comma = "";
        for (Column c : ts.getColumns()) {
            out.format("%s%s", comma, c.getName());
            comma = ",";
        }
        out.format("\n");
        for (Tuple tup : tuples()) { ///
            comma = "";
            for (Column k : tup.getColumns()) {
                String val = tup.get(k.getName()).trim();
                out.format("%s%s", comma, val);
                comma = ",";
            }
            out.format("\n");
        }
    }

    /**
     * convert tbl into a stream of Tuples
     *
     * @return stream of tbl tuples
     */
    public Stream<Tuple> stream() {
        return tuples().stream();
    }
}
