package CD;

import M4Parsing.LineParsers.CDLineParsers;
import M4Parsing.CDPrims.PrimTokens.Assoc;
import M4Parsing.CDPrims.PrimTokens.CD;
import M4Parsing.CDPrims.PrimTokens.Cons;
import Parsing.DBPrimitives.FixedStrTokens.AbsTaBle;
import Parsing.DBPrimitives.FixedStrTokens.DBase;
import Parsing.DBPrimitives.FixedStrTokens.OODB;
import Parsing.DBPrimitives.FixedStrTokens.SubTaBle;
import Parsing.DBPrimitives.FixedStrTokens.TaBle;
import Parsing.LineParsers.DBLineParsers;
import PrologDB.Column;
import PrologDB.DB;
import PrologDB.DBSchema;
import PrologDB.ErrorReport;
import PrologDB.OOSchema;
import PrologDB.SubTableSchema;
import PrologDB.Table;
import PrologDB.TableSchema;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.util.HashMap;
import java.util.LinkedList;
import M4Utilities.Error;
import M4Utilities.ParsE;
import M4Utilities.Utils;

/**
 * A class diagram schema
 *
 */
public class CDSpec extends OOSchema {

    public static final String cdspecEnd = ".cdspec.pl";

    protected LinkedList<ConsSchema> constraints;
    protected LinkedList<AssocSchema> associations;

    // CONSTRUCTORS
    /**
     * This constructor is used for the incremental construction of dbase or
     * oodb Schemas. Can incrementally add tables and then add subtable schemas.
     *
     *
     * @param name -- cdName given to schema
     */
    public CDSpec(String name) {
        super(name);
        this.name = name;
        this.constraints = new LinkedList<>();
        this.associations = new LinkedList<>();
    }

    public CDSpec union(CDSpec c) {
        this.setUnfinished();
        // Step 1: copy all new table/class definitions
        //         loop over the new stuff
        for (TableSchema ts : c.getTableSchemas()) {
            TableSchema tts = this.getTableSchema(ts.getName());
            if (tts == null) {
                this.add(ts);  // NOTE SHARING declarations now exists!
            } else {
                // c has a class with the same name.  add new columns to tts
                // loop over possible new fields/columns
                for (Column col : ts.getColumns()) {
                    Column k = tts.getColumn(col);
                    if (k == null) {
                        tts.add(col); // NOTE SHARING declarations now exists!
                    }
                }
            }
        }

        // Step 2: copy all new association definitions
        //         loop over new stuff
        for (AssocSchema as : c.getAssocSchemas()) {
            AssocSchema tas = this.getAssocSchema(as.getName());
            if (tas == null) {
                this.add(as); // NOTE SHARING declarations now exists!
            }
        }

        // Step 3: copy all subTableSchemas
        //         loop over new stuff
        for (SubTableSchema sts : c.getSubTableSchemas()) {
            SubTableSchema tsts = this.getSubTableSchema(sts.getName());
            if (tsts == null) {
                this.add(sts);  // NOTE SHARING declarations now exists!
            }
        }

        // Step 4: return modified 'this' CDSpec
        this.setFinished();
        return this;
        // write out CDSpecs and read them back in to eliminate 
        // structure sharing.!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    }

    /**
     * copyFrom an oodb schema.
     *
     * @return copied schema
     */
    public CDSpec copy() {
        CDSpec s = new CDSpec(name);
        s.copyFrom(this);
        return s;
    }

    /**
     * extension of equals
     *
     * @param cdspec an instance of CDSpec
     * @return true if both CDSpes are equal
     */
    public boolean equals(CDSpec cdspec) {
        super.equals(cdspec);

        if (associations.size() != cdspec.associations.size()) {
            return false;
        }

        System.err.println("CDSpec.equalsEH does not deal with constraints or equality of associations");

        return (listEquals(associations, cdspec.associations));
    }

    /**
     * two DBSchemas are equal iff (a) they have the same SubTableSchema, (b)
     * they have the same set of table schemas, and (c) they have the same set
     * of subtable schemas; if so this method returns. Otherwise a Error is
     * thrown
     *
     * @param schema.
     */
    public void equalsEH(CDSpec schema) {
        super.equalsEH(schema);

        if (!listEquals(associations, schema.associations)) {
            throw Error.toss(Error.associationsDiffer, schema.getName(), getName());
        }

        System.err.println("CDSpec.equalsEH does not deal with constraints or equality of associations");
    }

    /**
     * copyFrom core elements of a database or oodb schema
     *
     * @param s -- copy TableSchemas and SubTables from s into 'this'
     */
    public void copyFrom(CDSpec s) {
        super.copyFrom(s);
        associations.addAll(s.associations);
    }

    // ACCESSORS, SETTERS, and INFORMATION
    /**
     * a hack to set a schema that was formerly finished open to edits
     */
    public void setUnfinished() {
        finished = false;
    }

    /**
     * a hack to set a schema that was formerly finished closed to edits
     */
    public void setFinished() {
        finished = true;
    }

    /**
     * full (file) SubTableSchema of a schema is (SubTableSchema).schema.pl
     *
     * @return manufactured file SubTableSchema of schema
     */
    @Override
    public String getFullName() {
        return name + cdspecEnd;
    }

    /**
     * @param name of requested schema
     * @return requested association schema
     */
    public AssocSchema getAssocSchema(String name) {
        for (AssocSchema s : associations) {
            if (s.getName().equals(name)) {
                return s;
            }
        }
        return null;
    }

    /**
     * @param s -- AssocSchema to add
     */
    public void add(AssocSchema s) {
        if (getAssocSchema(s.getName()) != null) {
            throw new RuntimeException("multiple associations with name " + s.getName() + " declared");
        }
        associations.add(s);
    }

    /**
     * @param name of requested schema
     * @return requested association schema
     */
    public AssocSchema getAssocSchemaEH(String name) {
        AssocSchema result = getAssocSchema(name);
        if (result == null) {
            throw Error.toss(Error.associationNoExist, name);
        }
        return result;
    }
    
    public void remove(AssocSchema s) {
        associations.remove(s);
    }

    /**
     * @return list of collected AssocSchemas
     */
    public LinkedList<AssocSchema> getAssocSchemas() {
        return associations;
    }

    private boolean listEquals(LinkedList<AssocSchema> list1, LinkedList<AssocSchema> list2) {
        if (list1.size() != list2.size()) {
            return false;
        }
        for (AssocSchema outer : list1) {
            boolean ok = false;
            for (AssocSchema inner : list2) {
                if (outer.equals(inner)) {
                    ok = true;
                    break;
                }
            }
            if (ok == false) {
                return false;
            }
        }
        return true; // they're the same
    }

    // READ FILE UTILITIES
    /**
     * reads .ooschema; throws Error when an error is encountered. Returns
     * DBSchema declaration in file; may flatten it if .ooschema is read.
     *
     * @param schemafile -- SubTableSchema of schema file
     * @return DBSchema declaration of file
     */
    public static CDSpec read(String schemafile) {
        File in = new File(schemafile);
        return CDSpec.read(in);
    }

    /**
     * another version of readSchema
     *
     * @param schemafile -- File to read
     * @return DBSchema object
     */
    public static CDSpec read(File schemafile) {
        String filename = schemafile.getPath().replace("\\", "/");
        LineNumberReader br = readFileCheck(schemafile, filename);
        return read(filename, br);
    }

    /**
     * helper method -- does basic checks on File schemafile
     *
     * @param schemafile -- file that contains database schema
     */
    private static LineNumberReader readFileCheck(File schemafile, String filename) {
        LineNumberReader br;

        if (filename.endsWith(cdspecEnd)) {
            try {
                br = new LineNumberReader(new InputStreamReader(new FileInputStream(filename)));
                return br;
            } catch (FileNotFoundException e) {
                throw Error.toss(Error.schemaFileNoExist, filename);
            }
        }
        throw Error.toss(Error.wrongFileName, filename);
    }

    /**
     * used when it is easier to pass in a string declaration of a DBSchema
     * rather than building it programmatically
     *
     * @param nameOfSchema -- cdName of pseudo file
     * @param stringDefOfSchema -- string contents (typically multiline) of
     * pseudo file
     * @return DBSchema -- of pseudo file
     */
    public static CDSpec read(String nameOfSchema, String stringDefOfSchema) {
        InputStream is = new ByteArrayInputStream(stringDefOfSchema.getBytes());
        LineNumberReader br = new LineNumberReader(new InputStreamReader(is));
        return read(nameOfSchema, br);
    }

    /**
     * reads .cdschema.pl schema
     *
     * @param filename the value of filename
     * @param br - line number reader of file to read
     * @return the PrologDB.DBSchema
     */
    private static CDSpec read(String filename, LineNumberReader br) {
        CDSpec dbs = null;
        ErrorReport er = new ErrorReport();
        try {
            String line;
            while ((line = br.readLine()) != null) {
                line = line.trim();
                if (line.equals("")
                        || line.startsWith("%")
                        || line.startsWith("//")) {
                    continue;
                }
                int lineno = br.getLineNumber();

                try {
                    if (line.startsWith(CD.tok)) {
                        if (dbs != null) {
                            throw ParsE.toss(ParsE.dbDeclsTooMany, lineno, filename);
                        }
                        dbs = CDLineParsers.parseClassDiagramLine(line, lineno);
                    } else if (line.startsWith(TaBle.tok) || line.startsWith(AbsTaBle.tok)) {
                        DBLineParsers.parseTableLine(line, lineno, dbs, filename, null, er);
                    } else if (line.startsWith(SubTaBle.tok)) {
                        DBLineParsers.parseSubTableLine(line, lineno, dbs, filename, er);
                    } else if (line.startsWith(Assoc.tok)) {
                        CDLineParsers.parseAssocLine(line, lineno, dbs, er);
                    } else if (line.startsWith(DBase.tok) || line.startsWith(OODB.tok)) {
                        er.add("line %d: inappropriate to have a '%s' statement in an %s file", lineno, line, CD.tok);
                    } else if (line.startsWith(Cons.tok)) {
                        CDLineParsers.parseConsLine(line, lineno, dbs, er);
                    } else {
                        er.add("line %d: cannot parse '%s'", lineno, line);
                    }
                } catch (RuntimeException e) {
                    throw ParsE.toss(ParsE.tableSchemaParseError, br.getLineNumber(), filename, e.getMessage());
                }
            }
            br.close();

            // now basic verification.  tableNames is a list of names
            // produced when parsing a dbase declaration.  make sure that
            // schemas for these tables (and no others) have been created
            if (dbs == null) {
                er.add("no %s statement in this ooschema", CD.tok);
            }

            // Finally, no table should have an attribute with cdName "id"
            // as this is a reserved cdName (and is added to dominant tables
            // when translating to a cd form.
            for (TableSchema ts : dbs.getTableSchemas()) {
                for (Column c : ts.getColumns()) {
                    if (c.getName().equals("id")) {
                        er.add("table %s has column with reserved name 'id'", ts.getName());
                    }
                }
            }
        } catch (IOException e) {
            throw Error.toss(Error.ioerror, filename, e.getMessage());
        }
        String a[] = Utils.parseFileName(filename);
        String AppName = a[2];
        if (!dbs.getName().equals(AppName)) {
            er.add("file %s should contain class diagram %s, not %s", filename, AppName, dbs.getName());
        }
        for (AssocSchema as : dbs.getAssocSchemas()) {
            String c1name = as.getClass1();
            String c2name = as.getClass2();
            String f1name = as.getRole1();
            String f2name = as.getRole2();
            CheckFieldNameNotTaken(dbs, as, c1name, f2name, er);
            CheckFieldNameNotTaken(dbs, as, c2name, f1name, er);
        }
        er.printEH(System.out);
        return dbs;

        // finally, depending on who is calling this method, the flatten attribute is to be set.
    }

    private static void CheckFieldNameNotTaken(CDSpec dbs, AssocSchema as, String className, String fieldName, ErrorReport er) {
        TableSchema ts = dbs.getTableSchema(className);
        if (ts == null) {
            er.add("class %s referenced in association %s doesn't exist", className, as.getName());
            return;
        }
        Column c = ts.getColumn(fieldName);
        if (c != null) {
            er.add("field %s of class %s cannot be overridden by association end with same name", fieldName, className);
        }
        return;
    }

    @Override
    public void print(String filename) {
        print(filename, cdspecEnd);
    }

    // printing CD schema
    @Override
    public String toString() {
        String result = String.format("%s %s.\n\n", CD.tok, name);
        for (TableSchema ts : tableSchemas) {
            result += ts.toString();
        }
        result += "\n";
        for (SubTableSchema sts : subtables) {
            result += sts.toString();
        }
        if (subtables.size() > 0) {
            result += "\n";
        }
        for (AssocSchema a : associations) {
            result += a.toString();
        }
        if (associations.size() > 0) {
            result += "\n";
        }
        for (ConsSchema c : constraints) {
            result += c.toString();
        }
        if (constraints.size() > 0) {
            result += "\n";
        }
        return result;
    }

    public static DB toCD(CDSpec cds) {
        return cds.toCD();
    }

    /**
     * A CD database is flat -- has no subtables/inheritance. BUT it does encode
     * inheritance in table sub -- no attriburtes are propagated reason:
     * associations are attributes, but they are not treated as fields until the
     * CD schema is converted into a dbase (schema).
     *
     * @return CD datbase
     */
    public DB toCD() {
        try {
            String cdschema = M4Utilities.Utils.M4Home() + "libpl/cd.schema.pl";
            DBSchema dbs = DBSchema.read(cdschema);
            DB db = new DB(name, dbs);
            Table tab = db.getTableEH("tab");
            Table attr = db.getTableEH("attr");
            Table assoc = db.getTableEH("assoc");
            Table sub = db.getTableEH("sub");
            Table cons = db.getTableEH("cons");

//            // Step 0: convert CDSchema -- which is an OOschema -- into   /// BIG CHANGE
//            // a DB schema
//            DBSchema cddbs = this.finishAndPropagateAttributes();
            // step 1: extract each table and get its attributes
            HashMap<String, String> lookup = new HashMap<>();
            for (TableSchema s : this.getTableSchemas()) {
                String tname = s.getName();
                String tid = tname;
                String b = s.isAbstract() ? "true" : "false";
                tab.add(tid, tname, b);
                lookup.put(tname, tid);
//                attr.add(tid + "_id", "id", "str", tid);   // Don removed in debugging cdspec-to-schema
                for (Column c : s.getColumns()) {
                    String cname = c.getName();
                    String cid = cname;
                    String kind = (c.isQuoted()) ? "qstr" : c.getKind();
                    attr.add(tid + "_" + cid, cname, kind, tid);
                }
            }

            // Step 2: now do the same for associations
            for (AssocSchema a : getAssocSchemas()) {
                String aname = a.getName();
                String class1 = lookup.get(a.getClass1());
                String class2 = lookup.get(a.getClass2());
                assoc.add(aname, aname, class1, a.getRole1(), a.getAt1().toString(), a.min1 + "", a.max1 + "",
                        class2, a.getRole2(), a.getAt2().toString(), a.min2 + "", a.max2 + "");
            }

            // step 3: extract subtable tuples
            for (SubTableSchema ss : getSubTableSchemas()) {
                String supname = ss.getSuper().getName();
                String supid = lookup.get(supname);
                for (TableSchema ts : ss.getSubTableSchemas()) {
                    String subname = ts.getName();
                    String subid = lookup.get(subname);
                    sub.add(supid + "_" + subid, supid, subid);
                }
            }

            // Step 4: now do the same for constraints
            for (ConsSchema c : getConsSchemas()) {
                String cname = c.getName();
                cons.add(cname, cname, c.getConstraint());
            }

            return db;

        } catch (RuntimeException e) {
            System.err.format("error in translating cdschema %s to CD\n", name);
            throw e;
        }
    }

    /**
     * @param s -- ConsSchema to add
     */
    public void add(ConsSchema s) {
        if (getConsSchema(s.getName()) != null) {
            throw new RuntimeException("Multiple constraints with name " + s.getName() + " declared");
        }
        constraints.add(s);
    }

    /**
     * @param name of requested schema
     * @return requested cons schema
     */
    public ConsSchema getConsSchema(String name) {
        for (ConsSchema s : constraints) {
            if (s.getName().equals(name)) {
                return s;
            }
        }
        return null;
    }

    /**
     * @param name of requested schema
     * @return requested association schema
     */
    public ConsSchema getConsSchemaEH(String name) {
        ConsSchema result = getConsSchema(name);
        if (result == null) {
            throw Error.toss(Error.consSchemaNoExist, name);
        }
        return result;
    }

    /**
     * @return list of collected AssocSchemas
     */
    public LinkedList<ConsSchema> getConsSchemas() {
        return constraints;
    }

    public void addId(TableSchema ts) {
        ts.addFirst("id");
    }

    /**
     * For use in projecting CDSpecs for AoclBET
     *
     * @param TorB -- "T" or "B"
     * @param CSVfile --- CSV Diagram Differentiator Table
     * @return projected CDSpec, a proper subset of 'this'
     */
    public CDSpec project(String TorB, Table CSVfile) {
        // Step 1: create target CDSpec
        String cdName = TorB.equals("T") ? "Top" : "Bottom";
        CDSpec target = new CDSpec(cdName);

        // Step 2: project tables with their fields
        for (TableSchema ts : getTableSchemas()) {
            String tsName = ts.getName();
            if (containsTable(TorB, tsName, CSVfile)) {
                TableSchema tsnew = new TableSchema(tsName);
                for (Column c : ts.getColumns()) {
                    String fName = c.getName();
                    if (containsField(TorB, tsName, fName, CSVfile)) {
                        tsnew.add(c);
                    }
                }
                target.add(tsnew);
            }
        }

        // Step 3: project associations
        for (AssocSchema as : getAssocSchemas()) {
            String asName = as.getName();
            if (containsAssoc(TorB, asName, CSVfile)) {
                target.add(as);
            }
        }

        // Step 4: project inheritance
        for (SubTableSchema ss : getSubTableSchemas()) {
            TableSchema sup = ss.getSuper();
            String supName = sup.getName();
            if (containsTable(TorB, supName, CSVfile)) {
                SubTableSchema ssnew = new SubTableSchema(sup);
                for (TableSchema sts : ss.getSubTableSchemas()) {
                    String subName = sts.getName();
                    if (containsTable(TorB, subName, CSVfile)) {
                        ssnew.add(sts);
                    }
                }
                target.add(ssnew);
            }
        }

        // Step 5: return target CDSchema
        return target;
    }

    private boolean containsTable(String TorB, String cname, Table CSVfile) {
//        CSVfile.print();  // debugging
        Table result = CSVfile.select(c -> c.is("TorB", TorB) && c.is("CorA", "C") && c.is("name", cname) && c.is("field", ""));
        return result.isEmpty();
    }

    private boolean containsField(String TorB, String cname, String fname, Table CSVfile) {
        boolean result = CSVfile.select(c -> c.is("TorB", TorB) && c.is("CorA", "C") && c.is("field", fname) && c.is("name", cname)).isEmpty();
        return result;
    }

    private boolean containsAssoc(String TorB, String aname, Table CSVfile) {
        boolean result = CSVfile.select(c -> c.is("TorB", TorB) && c.is("CorA", "A") && c.is("name", aname)).isEmpty();
        return result;
    }
    
    // calling sequence   <inName.cdspec.pl> <histName.cdspec.pl> <unionName.cdspec.pl> args
    public static void Union(String inName, String histName, String unionName) {
        // Step 0: standard marquee processing
        if (!inName.endsWith(cdspecEnd) || !histName.endsWith(cdspecEnd) || !unionName.endsWith(cdspecEnd)) {
            System.out.format("One or more input files do not end in %s\n", cdspecEnd);
            System.exit(1);
        }
        if (inName.equals(unionName) || histName.equals(unionName)) {
            System.out.format("No overriding of input file (%s or %s) with output file (%s) permitted\n", inName,histName,unionName);
            System.exit(1);
        }
        CDSpec in = CDSpec.read(inName);
        CDSpec hs = CDSpec.read(histName);
        CDSpec newhs = hs.union(in);
        String newName = MDLUtilities.Utils.getAppName(unionName, cdspecEnd);
        newhs.rename(newName);
        newhs.print(unionName);
    }
}
